"""

Classes that manage hardware logic.

Each hardware class should be able to operate independently - ie. not
be dependent on a particular task class, etc. Other than that there are
very few design requirements:

* Every class should have a .release() method that releases any system
  resources in use by the object, eg. objects that use pigpio must have
  their `pigpio.pi` client stopped; LEDs should be explicitly turned off.
* The very minimal class attributes are described in the :class:`Hardware` metaclass.
* Hardware methods are typically called in their own threads, so care should
  be taken to make any long-running operations internally threadsafe.

.. _numbering-note:

Note:
    This software was primarily developed for the Raspberry Pi, which
    has `two types of numbering schemes <https://pinout.xyz/#>`_ ,
    "board" numbering based on physical position (e.g. pins 1-40, in 2 rows of 20 pins) and "bcm" numbering
    based on the broadcom chip numbering scheme (e.g. GPIO2, GPIO27).

    Board numbering is easier to use, but `pigpio <http://abyz.me.uk/rpi/pigpio/>`_
    , which we use as a bridge between Python and the GPIOs, uses the BCM scheme.
    As such each class that uses the GPIOs takes a board number as its argument
    and converts it to a BCM number in the __init__ method.

    If there is sufficient demand to make this more flexible, we can implement
    an additional `pref` to set the numbering scheme, but the current solution
    works without getting too muddy.

Warning:
    In order to use pigpio, the pigpio daemon must be running. See `the docs <http://abyz.me.uk/rpi/pigpio/python.html>`_
    Usually :class:`~.core.pilot.Pilot` s should be started by the bash script or systemd service
    generated by :mod:`.setup.setup_pilot`, which starts pigpiod.
"""

import typing
import warnings
import json
from pathlib import Path
import importlib

from autopilot import prefs
from autopilot.networking import Net_Node
from autopilot.core.loggers import init_logger
from autopilot.utils.common import NumpyEncoder, NumpyDecoder

# FIXME: Hardcoding names of metaclasses, should have some better system of denoting which classes can be instantiated
# directly for setup and prefs management.
META_CLASS_NAMES = ['Hardware', 'Camera', 'GPIO', 'Directory_Writer', 'Video_Writer']

# pigpio only uses BCM numbers, we need to translate them
# See https://www.element14.com/community/servlet/JiveServlet/previewBody/73950-102-11-339300/pi3_gpio.png
BOARD_TO_BCM = {
     3: 2,   5: 3,   7: 4,   8: 14, 10: 15,
    11: 17, 12: 18, 13: 27, 15: 22, 16: 23,
    18: 24, 19: 10, 21: 9,  22: 25, 23: 11,
    24: 8,  26: 7,  29: 5,  31: 6,  32: 12,
    33: 13, 35: 19, 36: 16, 37: 26, 38: 20,
    40: 21
}
"""
dict: Mapping from board (physical) numbering to BCM numbering. 

See `this pinout <https://pinout.xyz/#>`_.

Hardware objects take board numbered pins and convert them to BCM 
numbers for use with `pigpio`.
"""

BCM_TO_BOARD = dict([reversed(i) for i in BOARD_TO_BCM.items()])
"""
dict: The inverse of :const:`BOARD_TO_BCM`.
"""

class Hardware(object):
    """
    Generic class inherited by all hardware. Should not be instantiated
    on its own (but it won't do anything bad so go nuts i guess).

    Primarily for the purpose of defining necessary attributes.

    Also defines `__del__` to call `release()` so objects are always released
    even if not explicitly.

    Attributes:
        name (str): unique name used to identify this object within its group.
        group (str): hardware group, corresponds to key in prefs.json ``"HARDWARE": {"GROUP": {"ID": {**params}}}``
        is_trigger (bool): Is this object a discrete event input device?
            or, will this device be used to trigger some event? If `True`,
            will be given a callback by :class:`.Task`, and :meth:`.assign_cb`
            must be redefined.
        pin (int): The BCM pin used by this device, or None if no pin is used.
        type (str): What is this device known as in `.prefs`? Not required.
        input (bool): Is this an input device?
        output (bool): Is this an output device?
    """
    # metaclass for hardware objects
    is_trigger = False
    pin = None
    type = "" # what are we known as in prefs?
    input = False
    output = False

    def __init__(self, name=None, group=None, **kwargs):
        if name:
            self.name = name
        else:
            try:
                self.name = self.get_name()
            except:
                warnings.warn('wasnt passed name and couldnt find from prefs for object: {}'.format(self.__str__))
                self.name = None
        self.group = group

        self._calibration = None

        self.logger = init_logger(self)  # type: logging.Logger
        self.listens = {}
        self.node = None

    def release(self):
        """
        Every hardware device needs to redefine `release()`, and must

        * Safely unload any system resources used by the object, and
        * Return the object to a neutral state - eg. LEDs turn off.

        When not redefined, a warning is given.
        """
        raise Exception('The release method was not overridden by the subclass!')

    def assign_cb(self, trigger_fn):
        """
        Every hardware device that is a :attr:`~Hardware.trigger` must redefine this
        to accept a function (typically :meth:`.Task.handle_trigger`) that
        is called when that trigger is activated.

        When not redefined, a warning is given.
        """
        if self.is_trigger:
            raise Exception("The assign_cb method was not overridden by the subclass!")

    def get_name(self):
        """
        Usually Hardware is only instantiated with its pin number,
        but we can get its name from prefs
        """

        # TODO: Unify identification of hardware types across prefs and hardware objects
        try:
            our_type = prefs.get('HARDWARE')[self.type]
        except KeyError:
            our_type = prefs.get('HARDWARE')[self.__class__.__name__]

        for name, pin in our_type.items():
            if self.pin == pin:
                return name
            elif isinstance(pin, dict):
                if self.pin == pin['pin']:
                    return name

    def init_networking(self, listens=None, **kwargs):
        """
        Spawn a :class:`.Net_Node` to :attr:`Hardware.node` for streaming or networked command

        Args:
            listens (dict): Dictionary mapping message keys to handling methods
            **kwargs: Passed to :class:`.Net_Node`

        Returns:

        """

        if not listens:
            listens = self.listens

        self.node = Net_Node(
            self.name,
            upstream=prefs.get('NAME'),
            port=prefs.get('MSGPORT'),
            listens=listens,
            instance=False,
            **kwargs
            #upstream_ip=prefs.get('TERMINALIP'),
            #daemon=False
        )

    @property
    def calibration(self) -> typing.Optional[dict]:
        """
        Calibration used by the hardware object.

        Attempt to read from ``prefs.get('CALIBRATIONDIR')/group.name.json`` , if :attr:`Hardware.group` is ``None``,
        attempt to read from ``prefs.get('CALIBRATIONDIR')/name.json``

        Setting the attribute (over)writes the calibration to disk as a `.json` file

        Will be different for each hardware type, subclasses should document this property separately (eg.
        by overwriting ``Hardware.calibration.__doc__``

        Returns:
            (dict): if calibration is found, a dictionary of calibration for each property. None if no calibration found
        """
        if self._calibration is None:
            # try and find calibration file
            cal_name = None
            if self.name is not None and self.group is not None:
                cal_name = ".".join([self.group, self.name])
            elif self.name is not None:
                cal_name = self.name
            else:
                self.logger.debug('Hardware object has no group or name, cant find calibration!')

            if cal_name is not None:
                cal_name += ".json"

                path = Path(prefs.get('CALIBRATIONDIR')) / cal_name
                if path.exists():
                    with open(path, 'r') as cal_f:
                        self._calibration = json.load(cal_f, cls=NumpyDecoder)
                    self.logger.info(f'Calibration loaded from {path}')
                else:
                    self.logger.debug(f"No calibration found at {path}!")

        return self._calibration

    @calibration.setter
    def calibration(self, calibration):
        if calibration is None:
            self._calibration = calibration
            return
        # write to file
        # try and find calibration file
        cal_name = None
        if self.name is not None and self.group is not None:
            cal_name = ".".join([self.group, self.name])
        elif self.name is not None:
            cal_name = self.name
        else:
            self.logger.exception('Hardware has no group or name, dont know where to write calibration! saving in attribute, but will be lost on close!')

        if cal_name is not None:
            cal_name += '.json'
            cal_fn = Path(prefs.get('CALIBRATIONDIR')) / cal_name
            with open(cal_fn, 'w') as cal_f:
                json.dump(calibration, cal_f, cls=NumpyEncoder)
            self.logger.info(f'Calibration saved to {cal_fn}: \n{calibration}')

        self._calibration = calibration

    def __del__(self):
        self.release()




"""Module for generating and playing sounds.

This module contains the following files:

    sounds.py : Defines classes for generating sounds
    jackclient.py : Define the interface to the jack client
    pyoserver.py : Defines the interface to the pyo server

The use of pyoserver is discouraged in favor of jackclient. This is
controlled by the pref `AUDIOSERVER`.
"""



"""
Timeseries transformations, filters, etc.
"""
import typing
from time import time
from collections import deque
from autopilot.transform.transforms import Transform
from scipy import signal
import numpy as np
from copy import copy, deepcopy

class Filter_IIR(Transform):
    """
    Simple wrapper around :func:`scipy.signal.iirfilter`

    Creates a streaming filter -- takes in single values, stores them, and uses them to filter future values.

    Args:
        ftype (str): filter type, see ``ftype`` of :func:`scipy.signal.iirfilter` for available filters
        buffer_size (int): number of samples to store when filtering
        coef_type ({'ba', 'sos'}): type of filter coefficients to use (see :func:`scipy.signal.sosfilt` and :func:`scipy.signal.lfilt`)
        axis (int): which axis to filter over? (default: 0 because when passing arrays to filter, want to filter samples over time)
        **kwargs: passed on to :func:`scipy.signal.iirfilter` , eg.

            * ``N`` - filter order
            * ``Wn`` - array or scalar giving critical frequencies
            * ``btype`` - type of band: ``['bandpass', 'lowpass', 'highpass', 'bandstop']``

    Attributes:
        coefs (np.ndarray): filter coefficients, depending on :attr:`.coef_type`
        buffer (collections.deque): buffer of stored values to filter
        coef_type (str): type of filter coefficients to use (see :func:`scipy.signal.sosfilt` and :func:`scipy.signal.lfilt`)
        axis (int): which axis to filter over? (default: 0 because when passing arrays to filter, want to filter samples over time)
        ftype (str): filter type, see ``ftype`` of :func:`scipy.signal.iirfilter` for available filters
    """

    def __init__(self, ftype="butter", buffer_size=256, coef_type='sos', axis=0, *args, **kwargs):
        super(Filter_IIR, self).__init__(*args, **kwargs)

        self.ftype = ftype
        self.coef_type = coef_type
        self.axis = axis
        self.coefs = signal.iirfilter(ftype=self.ftype, output=coef_type, **kwargs)
        self.buffer = deque(maxlen=buffer_size)

    def process(self, input:float):
        """
        Filter the new value based on the values stored in :attr:`.Filter.buffer`

        Args:
            input (float): new value to filter!

        Returns:
            float: the filtered value!
        """

        self.buffer.append(input)
        if self.coef_type == "ba":
            return signal.lfilter(self.coefs[0], self.coefs[1], self.buffer, axis=self.axis)[-1]
        elif self.coef_type == "sos":
            return signal.sosfilt(self.coefs, self.buffer, axis=self.axis)[-1]


class Gammatone(Transform):
    """
    Single gammatone filter based on :cite:`slaneyEfficientImplementationPattersonHoldsworth1997`

    Thin wrapper around :class:`scipy.signal.gammatone` !! (started rewriting this and realized they had made a legible
    version <3 ty scipy team, additional implementations in the references)

    Examples:

        .. plot::

            import matplotlib.pyplot as plt
            import numpy as np

            from autopilot.transform.timeseries import Gammatone
            gamma_1k = Gammatone(1000, fs=44100)
            gamma_15k = Gammatone(15000, fs=44100)

            signal = (np.random.rand(44100)-0.5)*2
            filt_1k = gamma_1k.process(signal)
            filt_15k = gamma_15k.process(signal)

            fig, ax = plt.subplots(1,3, subplot_kw={'aspect':'equal'})
            ax[0].set_title('White Noise')
            ax[0].specgram(signal, Fs=44100)

            ax[1].set_title('1kHz Gammatone Filter')
            ax[1].axes.yaxis.set_visible(False)
            ax[1].specgram(filt_1k, Fs=44100)

            ax[2].set_title('15kHz Gammatone Filter')
            ax[2].axes.yaxis.set_visible(False)
            ax[2].specgram(filt_15k, Fs=44100)

            plt.tight_layout()
            plt.show()


    References:

        * :cite:`slaneyEfficientImplementationPattersonHoldsworth1997`
        * `Brian2hears implementation <https://github.com/brian-team/brian2hears/blob/131fd6d86c3ec460c45b42ea9c2f3b62c62d0631/brian2hears/filtering/filterbanklibrary.py#L26>`_
        * `detly/gammatone <https://github.com/detly/gammatone>`_
    """
    def __init__(self,
                 freq:float,
                 fs:int,
                 ftype:str="iir",
                 filtfilt:bool=True,
                 order:int=None,
                 numtaps:int=None,
                 axis:int=-1,
                 **kwargs):
        """
        Args:
            freq (float): Center frequency of the filter in Hz
            fs (int): Sampling rate of the signal to process
            ftype (str): Type of filter to return from :func:`scipy.signal.gammatone`
            filtfilt (bool): If ``True`` (default), use :func:`scipy.signal.filtfilt`, else use :func:`scipy.signal.lfilt`
            order (int): From scipy docs: The order of the filter. Only used when ``ftype='fir'``.
                Default is 4 to model the human auditory system. Must be between
                0 and 24.
            numtaps (int): From scipy docs: Length of the filter. Only used when ``ftype='fir'``.
                Default is ``fs*0.015`` if `fs` is greater than 1000,
                15 if `fs` is less than or equal to 1000.
            axis (int): Axis of input signal to apply filter over (default ``-1``)
            **kwargs: passed to :func:`scipy.signal.filtfilt` or :func:`scipy.signal.lfilt`
        """

        self._filter_a = None
        self._filter_b = None

        self.freq = float(freq)
        self.fs = int(fs)
        self.ftype = str(ftype)
        self.axis = int(axis)
        self.filtfilt = bool(filtfilt)

        self.order = order
        self.numtaps = numtaps
        self.kwargs = kwargs

        self._init_arrays()

    def _init_arrays(self):
        self._filter_b, self._filter_a = signal.gammatone(
            self.freq, self.ftype, self.order, self.numtaps, self.fs
        )

    def process(self, input:typing.Union[np.ndarray, list]) -> np.ndarray:
        if self.filtfilt:
            return signal.filtfilt(self._filter_b, self._filter_a, input,
                                   axis=self.axis, **self.kwargs)
        else:
            return signal.lfilter(self._filter_b, self._filter_a, input,
                                  axis=self.axis, **self.kwargs)







class Kalman(Transform):
    """
    Kalman filter!!!!!

    Adapted from https://github.com/rlabbe/filterpy/blob/master/filterpy/kalman/kalman_filter.py
    simplified and optimized lovingly <3

    Each of the arrays is named with its canonical letter and a short description, (eg. the x_state
    vector ``x_state`` is ``self.x_state``

    Args:
        dim_state (int): Dimensions of the state vector
        dim_measurement (int): Dimensions of the measurement vector
        dim_control (int): Dimensions of the control vector

    Attributes:
        x_state (:class:`numpy.ndarray`): Current state vector
        P_cov (:class:`numpy.ndarray`): Uncertainty Covariance
        Q_proc_var (:class:`numpy.ndarray`): Process Uncertainty
        B_control (:class:`numpy.ndarray`): Control transition matrix
        F_state_trans (:class:`numpy.ndarray`): State transition matrix
        H_measure (:class:`numpy.ndarray`): Measurement function
        R_measure_var (:class:`numpy.ndarray`): Measurement uncertainty
        M_proc_measure_xcor (:class:`numpy.ndarray`): process-measurement cross correlation
        z_measure (:class:`numpy.ndarray`):
        K (:class:`numpy.ndarray`): Kalman gain
        y (:class:`numpy.ndarray`):
        S (:class:`numpy.ndarray`): System uncertainty
        SI (:class:`numpy.ndarray`): Inverse system uncertainty
        x_prior (:class:`numpy.ndarray`): State prior
        P_prior (:class:`numpy.ndarray`): Uncertainty prior
        x_post (:class:`numpy.ndarray`): State posterior probability
        P_post (:class:`numpy.ndarray`): Uncertainty posterior probability

    References:
        Roger Labbe. "Kalman and Bayesian Filters in Python" - https://github.com/rlabbe/Kalman-and-Bayesian-Filters-in-Python
        Roger Labbe. "FilterPy" - https://github.com/rlabbe/filterpy
    """

    def __init__(self, dim_state: int, dim_measurement: int = None, dim_control: int=0,
                 *args, **kwargs):
        super(Kalman, self).__init__(*args, **kwargs)

        self.dim_state = dim_state # type: int
        if dim_measurement is None:
            self.dim_measurement = self.dim_state # type: int
        else:
            self.dim_measurement = dim_measurement # type: int
        self.dim_control = dim_control # type: int

        # initialize arrays!!!
        self._init_arrays()


    def _init_arrays(self, state=None):
        """
        Initialize the arrays!
        """

        # State arrays
        if state is not None:
            # TODO: check it's the right shape
            self.x_state = state
        else:
            self.x_state = np.zeros((self.dim_state, 1))

        # initialize kalman arrays
        self.P_cov               = np.eye(self.dim_state)                           # uncertainty covariance
        self.Q_proc_var          = np.eye(self.dim_state)                           # process uncertainty
        self.B_control           = np.eye(self.dim_control)                         # control transition matrix
        self.F_state_trans       = np.eye(self.dim_state)                           # x_state transition matrix
        if self.dim_state == self.dim_measurement:
            self.H_measure = np.eye(self.dim_measurement)
        else:
            self.H_measure           = np.zeros((self.dim_measurement, self.dim_state)) # measurement function
        self.R_measure_var       = np.eye(self.dim_measurement)                     # measurement uncertainty
        self._alpha_sq           = 1.                                               # fading memory control
        self.M_proc_measure_xcor = np.zeros((self.dim_state, self.dim_measurement)) # process-measurement cross correlation
        self.z_measure           = np.array([[None] * self.dim_measurement]).T

        # gain and residual are computed during the innovation step. We
        # save them so that in case you want to inspect them for various
        # purposes
        self.K = np.zeros((self.dim_state, self.dim_measurement)) # kalman gain
        self.y = np.zeros((self.dim_measurement, 1))
        self.S = np.zeros((self.dim_measurement, self.dim_measurement)) # system uncertainty
        self.SI = np.zeros((self.dim_measurement, self.dim_measurement)) # inverse system uncertainty

        # identity matrix. Do not alter this.
        self._I = np.eye(self.dim_state)

        # these will always be a copy of x_state,P_cov after predict() is called
        self.x_prior = self.x_state.copy()
        self.P_prior = self.P_cov.copy()

        # these will always be a copy of x_state,P_cov after update() is called
        self.x_post = self.x_state.copy()
        self.P_post = self.P_cov.copy()


    def predict(self, u=None, B=None, F=None, Q=None):
        """
        Predict next x_state (prior) using the Kalman filter x_state propagation
        equations.

        Update our state and uncertainty priors, :attr:`.x_prior` and :attr:`.P_prior`

        Parameters
        ----------

        u : np.array, default 0
            Optional control vector.

        B : np.array(dim_state, dim_u), or None
            Optional control transition matrix; a value of None
            will cause the filter to use `self.B_control`.

        F : np.array(dim_state, dim_state), or None
            Optional x_state transition matrix; a value of None
            will cause the filter to use `self.F_state_trans`.

        Q : np.array(dim_state, dim_state), scalar, or None
            Optional process noise matrix; a value of None will cause the
            filter to use `self.Q_proc_var`.
        """

        if B is None:
            B = self.B_control
        if F is None:
            F = self.F_state_trans
        if Q is None:
            Q = self.Q_proc_var
        elif np.isscalar(Q):
            Q = np.eye(self.dim_state) * Q



        # x_state = Fx + Bu
        if B is not None and u is not None:
            # make sure control vector is column
            u = np.atleast_2d(u)
            if u.shape[1] > u.shape[0]:
                u = u.T
            self.x_state = np.dot(F, self.x_state) + np.dot(B, u)
        else:
            self.x_state = np.dot(F, self.x_state)

        # P_cov = FPF' + Q_proc_var
        self.P_cov = self._alpha_sq * np.dot(np.dot(F, self.P_cov), F.T) + Q

        # save prior
        np.copyto(self.x_prior, self.x_state)
        np.copyto(self.P_prior, self.P_cov)


    def update(self, z:np.ndarray, R=None, H=None) -> np.ndarray:
        """
        Add a new measurement (z_measure) to the Kalman filter.

        If z_measure is None, nothing is computed. However, x_post and P_post are
        updated with the prior (x_prior, P_prior), and self.z_measure is set to None.

        Args:
            z (:class:`numpy.ndarray`): measurement for this update. z_measure can be a scalar if dim_measurement is 1,
                otherwise it must be convertible to a column vector.

                If you pass in a value of H_measure, z_measure must be a column vector the
                of the correct size.
            R (:class:`numpy.ndarray`, int, None): Optionally provide R_measure_var to override the measurement noise for this
                one call, otherwise  self.R_measure_var will be used.
            H (:class:`numpy.ndarray`, None): Optionally provide H_measure to override the measurement function for this
                one call, otherwise self.H_measure will be used.
        """

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

        if z is None:
            self.z_measure = np.array([[None] * self.dim_measurement]).T
            np.copyto(self.x_post, self.x_state)
            np.copyto(self.P_post, self.P_cov)
            self.y = np.zeros((self.dim_measurement, 1))
            return

        if R is None:
            R = self.R_measure_var
        elif np.isscalar(R):
            R = np.eye(self.dim_measurement) * R

        if H is None:
            z = self._reshape_z(z, self.dim_measurement, self.x_state.ndim)
            H = self.H_measure

        # y = z_measure - Hx
        # error (residual) between measurement and prediction
        self.y = z - np.dot(H, self.x_state)

        # common subexpression for speed
        PHT = np.dot(self.P_cov, H.T)

        # S = HPH' + R_measure_var
        # project system uncertainty into measurement space
        self.S = np.dot(H, PHT) + R
        self.SI = np.linalg.inv(self.S)
        # K = PH'inv(S)
        # map system uncertainty into kalman gain
        self.K = np.dot(PHT, self.SI)

        # x_state = x_state + Ky
        # predict new x_state with residual scaled by the kalman gain
        self.x_state = self.x_state + np.dot(self.K, self.y)

        # P_cov = (I-KH)P_cov(I-KH)' + KRK'
        # This is more numerically stable
        # and works for non-optimal K vs the equation
        # P_cov = (I-KH)P_cov usually seen in the literature.

        I_KH = self._I - np.dot(self.K, H)
        self.P_cov = np.dot(np.dot(I_KH, self.P_cov), I_KH.T) + np.dot(np.dot(self.K, R), self.K.T)

        # save measurement and posterior x_state
        np.copyto(self.z_measure, z)
        np.copyto(self.x_post, self.x_state)
        np.copyto(self.P_post, self.P_cov)
        return self.x_state

    def _reshape_z(self, z, dim_z, ndim):
        """ ensure z is a (dim_z, 1) shaped vector"""

        z = np.atleast_2d(z)
        if z.shape[1] == dim_z:
            z = z.T

        if z.shape != (dim_z, 1):
            raise ValueError('z (shape {}) must be convertible to shape ({}, 1)'.format(z.shape, dim_z))

        if ndim == 1:
            z = z[:, 0]

        if ndim == 0:
            z = z[0, 0]

        return z

    def process(self, z, **kwargs):
        """
        Call predict and update, passing the relevant kwargs

        Args:
            z ():
            **kwargs ():

        Returns:
            np.ndarray: self.x_state
        """

        # prepare args for predict and call
        predict_kwargs = {k:kwargs.get(k, None) for k in ("u", "B", "F", "Q")}
        self.predict(**predict_kwargs)

        # same thing for update
        update_kwargs = {k: kwargs.get(k, None) for k in ('R', 'H')}
        return self.update(z, **update_kwargs)








    def residual_of(self, z):
        """
        Returns the residual for the given measurement (z_measure). Does not alter
        the x_state of the filter.
        """
        return z - dot(self.H_measure, self.x_prior)

    def measurement_of_state(self, x):
        """
        Helper function that converts a x_state into a measurement.

        Parameters
        ----------

        x : np.array
            kalman x_state vector

        Returns
        -------

        z_measure : (dim_measurement, 1): array_like
            measurement for this update. z_measure can be a scalar if dim_measurement is 1,
            otherwise it must be convertible to a column vector.
        """

        return dot(self.H_measure, x)

    @property
    def alpha(self):
        """
        Fading memory setting. 1.0 gives the normal Kalman filter, and
        values slightly larger than 1.0 (such as 1.02) give a fading
        memory effect - previous measurements have less influence on the
        filter's estimates. This formulation of the Fading memory filter
        (there are many) is due to Dan Simon [1]_.
        """
        return self._alpha_sq**.5

    @alpha.setter
    def alpha(self, value):
        if not np.isscalar(value) or value < 1:
            raise ValueError('alpha must be a float greater than 1')

        self._alpha_sq = value**2



class Integrate(Transform):
    def __init__(self, decay=1, dt_scale = False, *args, **kwargs):
        super(Integrate, self).__init__(*args, **kwargs)
        self._value = None
        self.decay = decay
        self.dt_scale = dt_scale

        self.last_time = None

    def process(self, input):
        if self._value is None:
            self._value = input
            self.last_time = time()

        else:
            if self.dt_scale:
                new_time = time()
                input *= new_time-self.last_time
                self.last_time = new_time

            self._value += input
            self._value *= self.decay

        return self._value.copy()





















"""
Tests that will hopefully become unnecessary as testing coverage increases -- just testing that we can import things that we expect to be able to!
"""

def test_import_stim():
    from autopilot.stim import Stim


"""
Registry for programmatic access to autopilot classes and plugins

When possible, rather than importing and using an object directly, access it
using the ``get`` methods in this module. This makes it possible for plugins to
be integrated across the system.

"""
import pdb
from enum import Enum
import multiprocessing as mp
from typing import TYPE_CHECKING, Union, Optional, List, Type
import inspect


from autopilot.utils.common import find_class, recurse_subclasses, list_classes
from autopilot.utils.plugins import import_plugins, _IMPORTED
from autopilot import prefs
from autopilot.core.loggers import init_logger
if TYPE_CHECKING:
    from autopilot.hardware import Hardware
    from autopilot.tasks import Task

_IMPORT_LOCK = mp.Lock()
"""
Lock to ensure that another call to :func:`.registry.get` doesn't call :func:`.utils.plugins.import_plugins` twice
"""


class REGISTRIES(str, Enum):
    """
    Types of registries that are currently supported,
    ie. the possible values of the first argument of :func:`.registry.get`

    Values are the names of the autopilot classes that are searched for
    inheriting classes, eg. ``HARDWARE == "autopilot.hardware.Hardware"`` for :class:`autopilot.Hardware`
    """
    HARDWARE = "autopilot.hardware.Hardware"
    TASK = "autopilot.tasks.Task"
    GRADUATION = "autopilot.tasks.graduation.Graduation"
    TRANSFORM = "autopilot.transform.transforms.Transform"
    CHILDREN = "autopilot.tasks.children.Child"
    SOUND = "autopilot.stim.sound.sounds.BASE_CLASS"

# TODO: Update return type hint when unified Autopilot Object class made
def get(base_class:Union[REGISTRIES,str, type],
        class_name:Optional[str]=None,
        plugins:bool=True,
        ast:bool=True,
        include_base:bool=False) -> Union[type, List[type]]:
    """
    Get an autopilot object.

    Args:
        base_class (:class:`.REGISTRIES`, str, type): Class to search its subclasses for the indicated object. One of the values in the :class:`.REGISTRIES` enum,
           or else one of its keys (eg. ``'HARDWARE'``). If given a full module.ClassName string
           (eg. ``"autopilot.tasks.Task"``) attempt to get the indicated object. If given an object,
           use that.
       class_name (str, None): Name of class that inherits from base_class that is to be returned.
           if ``None`` (default), return all found subclasses of ``base_class``
       plugins (bool): If ``True`` (default), ensure contents of PLUGINDIR are loaded (with :func:`~.utils.plugins.import_plugins`)
           and are included in results. If ``False``, plugins are not explicitly imported, but if any
           have been imported elsewhere, they will be included anyway because we can't control all the
           different ways to subclass in Python.
       ast (bool): If ``True`` (default), if an imported object isn't found that matches ``class_name``,
           parse the syntax trees of submodules of ``base_class`` with :func:`.utils.common.list_classes`
           without importing to try and find it. If a match is found, it is imported and checked whether
           or not it is indeed a subclass of the ``base_class``. if ``False``, do not parse ast trees
           (will miss any modules that aren't already imported).
       include_base (bool): If ``False`` (default), remove the ``base_class`` before returning

    Returns:
        Either the requested items, or a list of all the relevant items
    """
    logger = init_logger(module_name='registry', class_name='get')

    if isinstance(base_class, REGISTRIES):
        base_class = base_class.value
    elif isinstance(base_class, str) and base_class.upper() in REGISTRIES.__members__.keys():
        base_class = REGISTRIES[base_class.upper()].value
    elif base_class in REGISTRIES.__members__.values():
        # already have the value, which is what we use to get the object
        pass
    elif inspect.isclass(base_class):
        # fine! we'll just use it
        pass
    else:
        logger.warning(f'Attempting to get subclasses from {base_class}, but it isn\'t in the REGISTRIES enum.\nPossible Values:\n{REGISTRIES.__members__}')

    if inspect.isclass(base_class):
        cls = base_class
    else:
        # get the class indicated by the base_class string
        cls = find_class(base_class)

    # load the contents of the plugin directory if we are supposed to
    # and if we haven't yet
    if not plugins or prefs.get('AUTOPLUGIN') is False:
        # if given explicit negative override, dont do it
        pass
    else:
        with globals()['_IMPORT_LOCK']:
            if not _IMPORTED.value:
                _ = import_plugins()

    # find subclasses!
    subclasses = recurse_subclasses(cls)
    # if we have found one among the imported classes, return that
    if class_name is not None:
        for subclass in subclasses:
            if subclass.__name__ == class_name:
                logger.debug(f"Found {subclass} as subclass of {cls} in module {subclass.__module__}")
                return subclass

    # if we haven't found the class yet and are asked to parse the ast tree, do it!
    if ast:
        ast_subclasses = list_classes(cls.__module__)
        for subclass in ast_subclasses:
            # if we were asked to get a specific class...
            if class_name is not None:
                if subclass[0] == class_name:
                    # import it and check that it is a subclass!
                    imported_subclass = find_class(subclass[1])
                    if issubclass(imported_subclass, cls):
                        logger.debug(f"Found {imported_subclass} as subclass of {cls} in {subclass[1]}")
                        return imported_subclass
            else:
                # get and import everything!
                try:
                    imported_subclass = find_class(subclass[1])
                    if issubclass(imported_subclass, cls):
                        subclasses.append(imported_subclass)
                except ImportError as e:
                    logger.warning(f'Exception importing requested class {subclass[1]}, got exception:\n{e}')

    # if all the found classes were requested, return them.
    if class_name is None:
        if not include_base:
            subclasses = [sub for sub in subclasses if sub is not cls]
        return sorted(list(set(subclasses)),
                      key=lambda subclass: ''.join([
                          subclass.__module__,
                          subclass.__name__
                      ]))

    # if we've gotten this far then we haven't found it :(
    ex_text = f"Could not find subclass of {cls} with name {class_name}!"
    logger.exception(ex_text)
    raise ValueError(ex_text)


def get_names(base_class:Union[REGISTRIES,str, type],
              class_name:Optional[str]=None,
              plugins:bool=True,
              ast:bool=True,
              full_name:bool=False) -> List[str]:
    """
    :func:`~.registry.get` but return a list of object names instead of the objects themselves

    See :func:`~.registry.get` for documentation of base arguments.

    .. note::

        While technically you can call this function with a ``class_name``,
        by default ``[class_name] == get_names(base_class, class_name)``,
        but if ``full_name == False`` it could be used to get the fully
        qualified package.module name in a pretty roundabout way.

    Args:
        full_name (bool): if ``False`` (default), return just the class name.
            if ``True``, return the full ``package.subpackage.module.Class_Name`` name.

    Returns:
        List[str]: a list of names
    """
    gotten = get(base_class=base_class,
                 class_name=class_name,
                 plugins=plugins,
                 ast=ast)

    if not isinstance(gotten, list):
        gotten = [gotten]

    if full_name:
        return ['.'.join([cls.__module__, cls.__name__]) for cls in gotten]
    else:
        return [cls.__name__ for cls in gotten]


def get_hardware(class_name:Optional[str] = None, plugins:bool = True, ast:bool = True) -> Union[Type["Hardware"], List[Type["Hardware"]]]:
    """
    Get a hardware class by name.

    Alias for :func:`.registry.get`

    Args:
        class_name (str): Name of hardware class to get
        plugins (bool): If ``True`` (default) ensure plugins are loaded and return
            from them. see :func:`.registry.get` for more details about the
            behavior of this argument
        ast (bool): If ``True`` (default) parse the syntax tree of all modules
            within :mod:`~autopilot.hardware`. see :func:`.registry.get` for more details about the
            behavior of this argument

    Returns:
        :class:`~autopilot.hardware.Hardware`
    """
    return get(REGISTRIES.HARDWARE, class_name=class_name, plugins=plugins, ast=ast)


def get_task(class_name:Optional[str] = None, plugins:bool = True, ast:bool = True) -> Union[Type["Task"], List[Type["Task"]]]:
    """
    Get a task class by name.

    Alias for :func:`.registry.get`

    Args:
        class_name (str): Name of task class to get
        plugins (bool): If ``True`` (default) ensure plugins are loaded and return
            from them. see :func:`.registry.get` for more details about the
            behavior of this argument
        ast (bool): If ``True`` (default) parse the syntax tree of all modules
            within :mod:`~autopilot.tasks`. see :func:`.registry.get` for more details about the
            behavior of this argument

    Returns:
        :class:`~autopilot.tasks.Task`
    """
    return get(REGISTRIES.TASK, class_name=class_name, plugins=plugins, ast=ast)









import ast
import inspect
import os
import typing
from collections import OrderedDict as odict

import npyscreen as nps

import autopilot
from autopilot.prefs import _DEFAULTS, Scopes
from autopilot.setup.scripts import SCRIPTS

# --------------------------------------------------
# split up prefs
# --------------------------------------------------

BASE_PREFS = odict({
    k:v for k, v in _DEFAULTS.items() if v["scope"] == Scopes.COMMON
})
PILOT_PREFS = odict({
    k:v for k, v in _DEFAULTS.items() if v["scope"] == Scopes.PILOT
})
LINEAGE_PREFS = odict({
    k:v for k, v in _DEFAULTS.items() if v["scope"] == Scopes.LINEAGE
})
AUDIO_PREFS = odict({
    k:v for k, v in _DEFAULTS.items() if v["scope"] == Scopes.AUDIO
})
TERMINAL_PREFS = odict({
    k:v for k, v in _DEFAULTS.items() if v["scope"] == Scopes.TERMINAL
})
DIRECTORY_STRUCTURE = odict({
    k:v for k, v in _DEFAULTS.items() if v["scope"] == Scopes.DIRECTORY
})

AGENTS = ('TERMINAL', 'PILOT', 'CHILD')

# --------------------------------------------------
# base class
# --------------------------------------------------

class Autopilot_Form(nps.Form):
    """
    Base class for Autopilot setup forms

    Each subclass needs to override the :meth:`.create` method to add the graphical elements for the form,
    typically by using :meth:`.populate_form` with a standardized description from :mod:`autopilot.prefs` , and
    the NAME attribute. The DESCRIPTION attribute provides title text for the form
    """

    NAME = "" # type: str
    DESCRIPTION = "" # type: str

    def __init__(self, prefs:dict, *args, **kwargs):
        self.prefs = prefs
        self.exports = {} # put any programmatically declared prefs here.
        self.input = odict()
        self.depends = {}
        super(Autopilot_Form, self).__init__(*args, **kwargs)

        if 'params' in kwargs.keys():
            self.populate_form(kwargs['params'])

    def populate_dependencies(self, params):
        # first find any options that other options depend on
        for param_name, param in params.items():
            if 'depends' in param.keys():
                if isinstance(param['depends'], str):
                    depends_on = param['depends']
                    depend_value = True
                elif isinstance(param['depends'], (tuple, list)):
                    depends_on = param['depends'][0]
                    depend_value = param['depends'][1]

                if depends_on in self.depends.keys():
                    self.depends[depends_on].append((param_name, depend_value))
                else:
                    self.depends[depends_on] = [(param_name, depend_value)]

    def populate_form(self, params):

        # check for existing values in global prefs

        self.populate_dependencies(params)

        # create widgets depending on parameter type
        for param_name, param in params.items():
            if param['type'] == 'bool':
                widget = self.add(nps.CheckBox, name=param['text'])
            elif param['type'] == 'choice':
                if param_name in self.prefs.keys():
                    try:
                        default_ind = [param['choices'].index(self.prefs[param_name])]
                    except ValueError:
                        default_ind = [0]
                else:
                    try:
                        default = param['default']
                        default_ind = [param['choices'].index(default)]
                    except KeyError:
                        default_ind = [0]

                widget = self.add(nps.TitleSelectOne,
                                  name = param['text'],
                                  values = param['choices'],
                                  max_height=len(param['choices'])+1,
                                  value = default_ind,
                                  scroll_exit = True)
            else:
                # params that just need a textbox, will attempt to be coerced
                # in unfold_values with ast.literal_eval.
                # eg. type in ('str', 'int', 'list', 'float')
                # try to get default from prefs, otherwise use the hardcoded default if present. otherwise blank
                if param_name in self.prefs.keys():
                    default = self.prefs[param_name]
                else:
                    try:
                        default = param['default']
                    except KeyError:
                        default = ''

                widget = self.add(nps.TitleText, name=param['text'], value=str(default))

            # if this widget depends on another, initially make it uneditable
            if 'depends' in param.keys():
                widget.editable = False
                widget.color = 'NO_EDIT'
                widget.labelColor = 'NO_EDIT'


            if param_name in self.depends.keys():

                widget.when_value_edited = lambda pname=param_name: self.update_depends(pname)

            self.input[param_name] = widget

    def update_depends(self, param_name):

        #pdb.set_trace()

        # get value
        param_value = self.input[param_name].value
        if hasattr(self.input[param_name], 'values'):
            param_value = self.input[param_name].values[param_value[0]]

        if param_name in self.depends.keys():
            for dependent in self.depends[param_name]:
                if param_value == dependent[1]:
                    self.input[dependent[0]].editable = True
                    self.input[dependent[0]].color = 'DEFAULT'
                    self.input[dependent[0]].labelColor = 'LABEL'
                else:
                    self.input[dependent[0]].editable = False
                    self.input[dependent[0]].color = 'NO_EDIT'
                    self.input[dependent[0]].labelColor = 'NO_EDIT'

    def afterEditing(self):
        self.parentApp.next_form_in_path(self)

# --------------------------------------------------
# individual forms
# --------------------------------------------------

class Hardware_Form(nps.FormWithMenus, Autopilot_Form):
    NAME = "HARDWARE"
    DESCRIPTION = "Configure Hardware Objects"

    def __init__(self, *args, **kwargs):
        self.input = odict()
        self.altrely = 4
        super(Hardware_Form, self).__init__(*args, **kwargs)


    def create(self):
        self.add(nps.FixedText, value="Use the ctrl+X menu to add new hardware", editable=False, color="VERYGOOD")

        # hardware_objs = self.list_hardware()
        hardware_objs = autopilot.get_hardware()
        # make dict, grouping by module
        hardware_dict = {}
        for hw_class in hardware_objs:
            module_name = hw_class.__module__.split('.')[-1]
            if module_name not in hardware_dict.keys():
                hardware_dict[module_name] = []
            hardware_dict[module_name].append(hw_class)

        for module, hardware_classes in hardware_dict.items():

            mod_menu = self.add_menu(module)
            for class_name in hardware_classes:
                mod_menu.addItem(text=class_name.__name__,
                                 onSelect=self.add_hardware,
                                 arguments=[class_name])

    def add_hardware(self, class_name):
        #self.nextrely = 1
        self.DISPLAY()

        # import the class
        if isinstance(class_name, str):
            hw_class = autopilot.get_hardware(class_name)
        else:
            hw_class = class_name
            class_name = hw_class.__name__
        # get its parent classes (which includes class itself)
        hw_parents = inspect.getmro(hw_class)
        # get signatures for each
        # go in reverse order so top classes options come first
        sigs = []
        # list to keep track of parameter names to remove duplicates
        param_names = []
        for cls in reversed(hw_parents):
            # get signature
            sig = inspect.signature(cls)
            # get parameter names and defaults
            for param_name, param in sig.parameters.items():
                param_default = param.default
                if param_default == inspect._empty:
                    param_default = None
                if param_name in ('kwargs', 'args'):
                    continue

                if param_name not in param_names:
                    # check if we already have a parameter with this name,
                    # if we don't add it.
                    sigs.append((param_name, param_default))
                    param_names.append(param_name)

        module = hw_class.__module__.split('.')[-1]
        MODULE = module.upper()
        # create title and input widgets for arguments

        #pdb.set_trace()

        self.add(nps.FixedText,
                 value="{}.{}".format(module, class_name),
                 rely=self.altrely,
                 editable=False,
                 color="VERYGOOD")

        self.altrely+=1

        hw_widgets = {}
        hw_widgets['type'] = "{}.{}".format(module, class_name)
        for sig in sigs:
            if sig[1] is None:
                sig = (sig[0], '')

            hw_widgets[sig[0]] = self.add(nps.TitleText,
                                          name=sig[0],
                                          value=str(sig[1]),
                                          rely=self.altrely)

            self.altrely+=1
        self.altrely+=1

        if MODULE not in self.input.keys():
            self.input[MODULE] = []

        self.input[MODULE].append(hw_widgets)


class Agent_Form(nps.Form):
    NAME="AGENT"
    DESCRIPTION = "Select an Agent"

    def create(self):
        # self.input = odict({
        #     'AGENT': self.add(nps.TitleSelectOne, max_height=len(AGENTS)+1, value=[0,],
        #                       name="Select an Agent. If this is a Raspberry Pi you should select either 'PILOT' or 'CHILD', and if this is the computer you will be using to control Autopilot you should select 'TERMINAL'", values = AGENTS, scroll_exit=True)
        # })

        if os.path.exists(os.path.join(os.path.dirname(__file__),'welcome_msg.txt')):
            with open(os.path.join(os.path.dirname(__file__),'welcome_msg.txt'), 'r') as welcome_f:
                welcome = welcome_f.read()
                for line in welcome.split('\n'):self.add(nps.FixedText, value=line, editable=False)

        self.input = odict({
            'AGENT': self.add(nps.TitleSelectOne, max_height=len(AGENTS)+1, value=0,
                              name="Select an Autopilot Agent", values=AGENTS, scroll_exit=True)
        })

    def afterEditing(self):
        # terminal

        if self.input['AGENT'].value[0] == 0:
            agent_name = "TERMINAL"
        elif self.input['AGENT'].value[0] == 1:
            agent_name = 'PILOT'
        elif self.input['AGENT'].value[0] == 2:
            agent_name = 'CHILD'
        else:
            self.parentApp.setNextForm(None)
            return

        self.parentApp.agent = agent_name
        self.parentApp.path = self.parentApp.PATHS[agent_name]
        self.parentApp.next_form_in_path(self)


class Directory_Form(Autopilot_Form):
    NAME='DIRECTORIES'
    DESCRIPTION = "Configure directory structure"

    def create(self):
        self.add(nps.FixedText, value="Directory Structure", editable=False, color="VERYGOOD")
        self.populate_form(DIRECTORY_STRUCTURE)


class Common_Form(Autopilot_Form):
    NAME="COMMON"
    DESCRIPTION = "Configure common prefs"
    def create(self):
        self.add(nps.FixedText, value="Common Prefs", editable=False, color="VERYGOOD")
        self.populate_form(BASE_PREFS)


class Scripts_Form(Autopilot_Form):
    NAME="SCRIPTS"
    DESCRIPTION = "Choose environment configuration scripts to run"
    def create(self):
        self.populate_form(SCRIPTS)


class Pilot_Config_Form_1(Autopilot_Form):
    NAME="PILOT"
    DESCRIPTION = "Configure Pilot-specific prefs"
    def create(self):
        self.add(nps.FixedText, value='Pilot Prefs', editable=False, color="VERYGOOD")
        self.populate_form(PILOT_PREFS)
        self.add(nps.FixedText, value='Lineage Prefs', editable=False, color="VERYGOOD")
        self.populate_form(LINEAGE_PREFS)
        self.add(nps.FixedText, value='Audio Prefs', editable=False, color="VERYGOOD")
        self.populate_form(AUDIO_PREFS)


class Terminal_Form(Autopilot_Form):
    NAME="TERMINAL"
    DESCRIPTION = "Configure Terminal-specific prefs"
    def create(self):
        self.add(nps.FixedText, value='Terminal Prefs', editable=False, color="VERYGOOD")
        self.populate_form(TERMINAL_PREFS)

# --------------------------------------------------
# Combined app object
# --------------------------------------------------

class Autopilot_Setup(nps.NPSAppManaged):
    PATHS = {
        'TERMINAL': ['AGENT', 'DIRECTORIES', 'COMMON', 'TERMINAL', 'SCRIPTS'],
        'PILOT': ['AGENT', 'DIRECTORIES', 'COMMON', 'PILOT', 'HARDWARE', 'SCRIPTS']
    }
    """
    Allow different agents to have different paths through setup
    
    For each Agent, a list of forms to call in order using :meth:`.next_form_in_path`
    
    Setup always starts with the :class:`.Agent_Form` which populates the :attr:`.agent` and :attr:`.path` attributes
    """

    def __init__(self, prefs):
        super(Autopilot_Setup, self).__init__()
        self.prefs = prefs
        self.agent = "" # type: str
        self.forms = {} # type: typing.Dict[str, Autopilot_Form]
        self.path = []



    def onStart(self):
        """
        Add forms by gathering subclasses of :class:`.Autopilot_Form`

        Set the :class:`.Agent_Form` as the ``'MAIN'`` form -- opened first.
        """
        self.forms['AGENT'] = self.addForm('MAIN', Agent_Form, name="Select Agent", prefs=self.prefs)

        # then iterate through subclasses and add
        for form_class in Autopilot_Form.__subclasses__():
            self.forms[form_class.NAME] = self.addForm(form_class.NAME, form_class, name=form_class.DESCRIPTION, prefs=self.prefs)

    def next_form_in_path(self, calling_form: Autopilot_Form):
        """
        Call :meth:`.setNextForm` to set the next form in the :attr:`.path`
        :attr:`.path` and :attr:`.agent` will be set by :class:`.Agent_Form`

        Args:
            calling_form (:class:`.Autopilot_Form`): The currently active form
        """
        # path = self.PATHS[self.agent]
        next_ind = self.path.index(calling_form.NAME) + 1
        if next_ind >= len(self.path):
            self.setNextForm(None)
        else:
            self.setNextForm(self.path[next_ind])

    def unpack_prefs(self):
        """
        Unpack the prefs from the forms and return them complete

        Returns:
            dict - your prefs!
        """

        out_prefs = odict()
        for form_name in self.path:
            # skip scripts, it's not exactly prefs
            if form_name == 'SCRIPTS':
                continue

            new_prefs = odict({k: unfold_values(v) for k, v in self.forms[form_name].input.items()})

            # hardware needs a little extra unpacking
            if form_name == "HARDWARE":
                hardware = {}
                for hardware_group, hardware_list in new_prefs.items():
                    hardware[hardware_group] = {}
                    for hardware_config in hardware_list:
                        hardware[hardware_group][hardware_config['name']] = hardware_config
                out_prefs['HARDWARE'] = hardware
            else:
                out_prefs.update(new_prefs)

        return out_prefs

    @property
    def active_scripts(self) -> typing.List[str]:
        """
        Get list of scripts that were selected to be run

        Returns:
            list:  script_name: script_commands
        """
        out_scripts = []

        script_names = odict({k: unfold_values(v) for k, v in self.forms['SCRIPTS'].input.items()})
        for script_name, result in script_names.items():
            if script_name in SCRIPTS.keys() and result == 1:
                out_scripts.append(script_name)

        return out_scripts


def unfold_values(v):
    """
    Unfold nested values from the SetupForm. Called recursively.

    Args:
        v (dict): unfolded values
    """
    if isinstance(v, dict):
        # recurse
        v = {k: unfold_values(v) for k, v in v.items()}
    elif isinstance(v, list):
        v = [unfold_values(v) for v in v]
    elif isinstance(v, str):
        # do nothing since this is what we want yno
        pass
    else:
        #pdb.set_trace()

        if hasattr(v, 'values'):
            # if it's an object that has mutiple values, ie. a choice box, value is inside a list.
            v = v.values[v.value[0]]
        elif hasattr(v, 'value'):
            # if it isn't a list, but is still a widget object, get the value
            v = v.value


        try:
            # convert ints to ints, lists to lists, etc. from strings
            v = ast.literal_eval(v)
        except:
            # fine, just a string that can't be evaluated into another type
            pass
    return v




"""
Client that dumps samples directly to the jack client with the :mod:`jack` package.
"""
from itertools import cycle
import multiprocessing as mp
import queue as queue
import numpy as np
from copy import copy
from queue import Empty


# importing configures environment variables necessary for importing jack-client module below
from autopilot import external
from autopilot.core.loggers import init_logger

try:
    import jack
except (OSError, ModuleNotFoundError):
    print('jack library not found! sounds unavailable')

from autopilot import prefs

# allows us to access the audio server and some sound attributes
SERVER = None
"""
:class:`.JackClient`: After initializing, JackClient will register itself with this variable.
"""

FS = None
"""
int: Sampling rate of the active server
"""

BLOCKSIZE = None
"""
int: Blocksize, or the amount of samples processed by jack per each :meth:`.JackClient.process` call.
"""

QUEUE = None
"""
:class:`multiprocessing.Queue`: Queue to be loaded with frames of BLOCKSIZE audio.
"""

PLAY = None
"""
:class:`multiprocessing.Event`: Event used to trigger loading samples from `QUEUE`, ie. playing.
"""

STOP = None
"""
:class:`multiprocessing.Event`: Event that is triggered on the end of buffered audio.

Note:
    NOT an event used to stop audio.
"""

Q_LOCK = None
"""
:class:`multiprocessing.Lock`: Lock that enforces a single writer to the `QUEUE` at a time.
"""

CONTINUOUS = None
"""
:class:`multiprocessing.Event`: Event that (when set) signals the sound server should play some sound continuously rather than remain silent by default (eg. play a background sound).

"""

CONTINUOUS_QUEUE = None
"""
:class:`multiprocessing.Queue`: Queue that 
"""

CONTINUOUS_LOOP = None
"""
:class:`multiprocessing.Event`: Event flag that is set when frames dropped into the CONTINUOUS_QUEUE should be looped (eg. in the case of stationary background noise),
otherwise they are played and then discarded (ie. the sound is continuously generating and submitting samples)
"""

class JackClient(mp.Process):
    """
    Client that dumps frames of audio directly into a running jackd client.

    When first initialized, sets module level variables above.

    Attributes:
        name (str): name of client, default "jack_client"
        q (:class:`~.multiprocessing.Queue`): Queue that stores buffered frames of audio
        q_lock (:class:`~.multiprocessing.Lock`): Lock that manages access to the Queue
        play_evt (:class:`multiprocessing.Event`): Event used to trigger loading samples from `QUEUE`, ie. playing.
        stop_evt (:class:`multiprocessing.Event`): Event that is triggered on the end of buffered audio.
        quit_evt (:class:`multiprocessing.Event`): Event that causes the process to be terminated.
        client (:class:`jack.Client`): Client to interface with jackd
        blocksize (int): The blocksize - ie. samples processed per :meth:`.JackClient.process` call.
        fs (int): Sampling rate of client
        zero_arr (:class:`numpy.ndarray`): cached array of zeroes used to fill jackd pipe when not processing audio.
        continuous_cycle (:class:`itertools.cycle`): cycle of frames used for continuous sounds
        mono_output (bool): ``True`` or ``False`` depending on if the number of output channels is 1 or >1, respectively.
            detected and set in :meth:`.JackClient.boot_server` , initialized to ``True`` (which is hopefully harmless)
    """
    def __init__(self, name='jack_client'):
        """
        Args:
            name:
        """
        super(JackClient, self).__init__()

        # TODO: If global client variable is set, just return that one.

        self.name = name
        #self.pipe = pipe
        self.q = mp.Queue()
        self.q_lock = mp.Lock()

        self.play_evt = mp.Event()
        self.stop_evt = mp.Event()
        self.quit_evt = mp.Event()

        # we make a client that dies now so we can stash the fs and etc.
        self.client = jack.Client(self.name)
        self.blocksize = self.client.blocksize
        self.fs = self.client.samplerate
        self.zero_arr = np.zeros((self.blocksize,1),dtype='float32')

        # a few objects that control continuous/background sound.
        # see descriptions in module variables
        self.continuous = mp.Event()
        self.continuous_q = mp.Queue()
        self.continuous_loop = mp.Event()
        self.continuous_cycle = None
        self.continuous.clear()
        self.continuous_loop.clear()

        # store the frames of the continuous sound and cycle through them if set in continous mode
        self.continuous_cycle = None

        # Something calls process() before boot_server(), so this has to
        # be initialized
        self.mono_output = True

        # store a reference to us and our values in the module
        globals()['SERVER'] = self
        globals()['FS'] = copy(self.fs)
        globals()['BLOCKSIZE'] = copy(self.blocksize)
        globals()['QUEUE'] = self.q
        globals()['Q_LOCK'] = self.q_lock
        globals()['PLAY'] = self.play_evt
        globals()['STOP'] = self.stop_evt
        globals()['CONTINUOUS'] = self.continuous
        globals()['CONTINUOUS_QUEUE'] = self.continuous_q
        globals()['CONTINUOUS_LOOP'] = self.continuous_loop

        self.logger = init_logger(self)

        if self.fs != prefs.get('FS'):
            self.logger.warning(
                f"Sampling rate was set to {prefs.get('FS')} in prefs, but the jack audio daemon is running at {self.fs}. \
                Check that jackd was not already running, and is being correctly started by autopilot (see autopilot.external)")
    

    def boot_server(self):
        """
        Called by :meth:`.JackClient.run` to boot the server upon starting the process.

        Activates the client and connects it to the physical speaker outputs
        as determined by `prefs.get('OUTCHANNELS')`.

        This is the interpretation of OUTCHANNELS:
        * empty string
            'mono' audio: the same sound is always played to all channels. 
            Connect a single virtual outport to every physical channel.
            If multi-channel sound is provided, raise an error.
        * a single int (example: J)
            This is equivalent to [J].
            The first virtual outport will be connected to physical channel J.
            Note this is NOT the same as 'mono', because only one speaker
            plays, instead of all speakers.
        * a list (example: [I, J])
            The first virtual outport will be connected to physical channel I.
            The second virtual outport will be connected to physical channel J.
            And so on.    
            If 1-dimensional sound is provided, play the same to all speakers
            (like mono mode).
            If multi-channel sound is provided and the number of channels
            is different form the length of this list, raise an error.        

        :class:`jack.Client` s can't be kept alive, so this must be called just before
        processing sample starts.
        """
        ## Parse OUTCHANNELS into listified_outchannels and set `self.mono_output`
        # Get the pref
        outchannels = prefs.get('OUTCHANNELS')
        
        # This generates `listified_outchannels`, which is always a list
        # It also sets `self.mono_output` if outchannels is None
        if outchannels == '':
            # Mono mode
            listified_outchannels = []
            self.mono_output = True
        elif not isinstance(outchannels, list):
            # Must be a single integer-like thing
            listified_outchannels = [int(outchannels)]
            self.mono_output = False
        else:
            # Already a list
            listified_outchannels = outchannels
            self.mono_output = False
        
        ## Initalize self.client
        # Initalize a new Client and store some its properties
        # I believe this is how downstream code knows the sample rate
        self.client = jack.Client(self.name)
        self.blocksize = self.client.blocksize
        self.fs = self.client.samplerate
        
        # This is used for writing silence
        self.zero_arr = np.zeros((self.blocksize,1),dtype='float32')

        # Set the process callback to `self.process`
        # This gets called on every chunk of audio data
        self.client.set_process_callback(self.process)

        # Register virtual outports
        # This is something we can write data into
        if self.mono_output:
            # One single outport
            self.client.outports.register('out_0')
        else:
            # One outport per provided outchannel
            for n in range(len(listified_outchannels)):
                self.client.outports.register('out_{}'.format(n))

        # Activate the client
        self.client.activate()
        
        
        ## Hook up the outports (data sinks) to physical ports
        # Get the actual physical ports that can play sound
        target_ports = self.client.get_ports(
            is_physical=True, is_input=True, is_audio=True)

        # Depends on whether we're in mono mode
        if self.mono_output:
            ## Mono mode
            # Hook up one outport to all channels
            for target_port in target_ports:
                self.client.outports[0].connect(target_port)
        
        else:
            ## Not mono mode
            # Error check
            if len(listified_outchannels) > len(target_ports):
                raise ValueError(
                    "cannot connect {} ports, only {} available".format(
                    len(listified_outchannels),
                    len(target_ports),))
            
            # Hook up one outport to each channel
            for n in range(len(listified_outchannels)):
                # This is the channel number the user provided in OUTCHANNELS
                index_of_physical_channel = listified_outchannels[n]
                
                # This is the corresponding physical channel
                # I think this will always be the same as index_of_physical_channel
                physical_channel = target_ports[index_of_physical_channel]
                
                # Connect virtual outport to physical channel
                self.client.outports[n].connect(physical_channel)

    def run(self):
        """
        Start the process, boot the server, start processing frames and wait for the end.
        """
        self.logger = init_logger(self)
        self.boot_server()

        # we are just holding the process open, so wait to quit
        try:
            self.quit_evt.clear()
            self.quit_evt.wait()
        except KeyboardInterrupt:
            # just want to kill the process, so just continue from here
            pass

    def quit(self):
        """
        Set the :attr:`.JackClient.quit_evt`
        """
        self.quit_evt.set()

    def process(self, frames):
        """
        Process a frame of audio.

        If the :attr:`.JackClient.play_evt` is not set, fill port buffers with zeroes.

        Otherwise, pull frames of audio from the :attr:`.JackClient.q` until it's empty.

        When it's empty, set the :attr:`.JackClient.stop_evt` and clear the :attr:`.JackClient.play_evt` .

        Args:
            frames: number of frames (samples) to be processed. unused. passed by jack client
        """
        ## Switch on whether the play event is set
        if not self.play_evt.is_set():
            # A play event has not been set
            # Play only if we are in continuous mode, otherwise write zeros
            
            ## Switch on whether we are in continuous mode
            if self.continuous.is_set():
                # We are in continuous mode, keep playing
                if self.continuous_cycle is None:
                    try:
                        to_cycle = self.continuous_q.get_nowait()
                        self.continuous_cycle = cycle(to_cycle)
                        self.logger.debug(f'started playing continuous sound with length {len(to_cycle)} frames')
                    except Empty:
                        self.logger.exception('told to play continuous sound but nothing in queue, will try again next loop around')
                        self.client.outports[0].get_array()[:] = self.zero_arr.T
                        return

                # Get the data to play
                data = next(self.continuous_cycle).T
                
                # Write
                self.write_to_outports(data)

            else:
                # We are not in continuous mode, play silence
                # clear continuous sound after it's done
                if self.continuous_cycle is not None:
                    self.continuous_cycle = None

                # Play zeros
                data = np.zeros(self.blocksize, dtype='float32')
                
                # Write
                self.write_to_outports(data)

        else:
            # A play event has been set
            # Play a sound

            # Try to get data
            try:
                data = self.q.get_nowait()
            except queue.Empty:
                data = None
                self.logger.warning('Queue Empty')
            
            
            ## Switch on whether data is available
            if data is None:
                # fill with continuous noise
                if self.continuous.is_set():
                    try:
                        data = next(self.continuous_cycle)
                    except Exception as e:
                        self.logger.exception(f'Continuous mode was set but got exception with continuous queue:\n{e}')
                        data = self.zero_arr

                else:
                    # Play zeros
                    data = np.zeros(self.blocksize, dtype='float32')


                # Write data
                self.write_to_outports(data)
                
                # sound is over
                self.play_evt.clear()
                self.stop_evt.set()
                
            else:
                ## There is data available

                # pad if needed
                if data.shape[0] < self.blocksize:
                    data = self.pad(data)

                # Write
                self.write_to_outports(data)

    def pad(self, data: np.ndarray) -> np.ndarray:
        """
        Pad a sound that is not a full chunk length long, either by filling with silence
        or with a continuous sound, if any.

        Args:
            data (:class:`numpy.ndarray`): The sound to pad!

        Returns:
            :class:`numpy.ndarray` - the padded sound!
        """
        # if sound was not padded, fill remaining with continuous sound or silence
        n_from_end = self.blocksize - data.shape[0]
        if self.continuous.is_set():
            # data = np.concatenate((data, self.continuous_cycle.next()[-n_from_end:]),
            #                       axis=0)
            try:
                cont_data = next(self.continuous_cycle)
                data = np.concatenate((data, cont_data[-n_from_end:]),
                                      axis=0)
            except Exception as e:
                self.logger.exception(f'Continuous mode was set but got exception with continuous queue:\n{e}')
                data = np.pad(data, (0, n_from_end), 'constant')
        else:
            data = np.pad(data, (0, n_from_end), 'constant')

        return data

    
    def write_to_outports(self, data):
        """Write the sound in `data` to the outport(s).
        
        If self.mono_output:
            If data is 1-dimensional:
                Write that data to the single outport, which goes to all
                speakers.
            Otherwise, raise an error.
        
        If not self.mono_output:
            If data is 1-dimensional:
                Write that data to every outport
            If data is 2-dimensional:
                Write one column to each outport, raising an error if there
                is a different number of columns than outports.
        """
        ## Write the output to each outport
        if self.mono_output:
            ## Mono mode - Write the same data to all channels
            if data.ndim == 1:
                # Write data to one outport, which is hooked up to all channels
                buff = self.client.outports[0].get_array()
                buff[:] = data
            
            else:
                # Stereo data provided, this is an error
                raise ValueError(
                    "pref OUTCHANNELS indicates mono mode, but "
                    "data has shape {}".format(data.shape))
            
        else:
            ## Multi-channel mode - Write a column to each channel
            if data.ndim == 1:
                ## 1-dimensional sound provided
                # Write the same data to each channel
                for outport in self.client.outports:
                    buff = outport.get_array()
                    buff[:] = data
                
            elif data.ndim == 2:
                ## Multi-channel sound provided
                # Error check
                if data.shape[1] != len(self.client.outports):
                    raise ValueError(
                        "data has {} channels "
                        "but only {} outports in pref OUTCHANNELS".format(
                        data.shape[1], len(self.client.outports)))
                
                # Write one column to each channel
                for n_outport, outport in enumerate(self.client.outports):
                    buff = outport.get_array()
                    buff[:] = data[:, n_outport]
                
            else:
                ## What would a 3d sound even mean?
                raise ValueError(
                    "data must be 1 or 2d, not {}".format(data.shape))




import pytest
import warnings

from autopilot import prefs

@pytest.mark.parametrize('default_pref', [(k, v) for k, v in prefs._DEFAULTS.items()])
def test_prefs_defaults(default_pref):

    # make sure that we didnt' actually load anything from some phantom uh prefs file idk

    # if prefs._INITIALIZED.value:
    #     warnings.warn('prefs was initialized, so defaults could not be tested')
    #     return
    # save any existing set prefs to restore at the end
    existing_prefs = prefs._PREFS._getvalue().copy()
    prefs._PREFS = prefs._PREF_MANAGER.dict()

    if 'default' in default_pref[1].keys():
        with pytest.warns(UserWarning):
            assert prefs.get(default_pref[0]) == default_pref[1]['default']
    else:
        assert prefs.get(default_pref[0]) is None

    # restore
    for k, v in existing_prefs.items():
        prefs._PREFS[k] = v


def test_prefs_deprecation():
    """
    If there is a string in the ``'deprecation'`` field of a pref in `_DEFAULTS`,
    a warning is raised printing the string.
    """

    # add a fake deprecated pref
    prefs._DEFAULTS['DEPRECATEME'] = {
        'type': 'int',
        "text": "A pref that was born just to die",
        "default": 4,
        "scope": prefs.Scopes.COMMON,
        'deprecation': 'This pref will be DECIMATED i mean DEPRECATED in a future version'
    }

    with pytest.warns(FutureWarning):
        pref_val = prefs.get('DEPRECATEME')

    assert pref_val == 4



"""
After initial setup, configure autopilot: create an autopilot directory and a prefs.json file

"""

import npyscreen as nps
import datetime
import _curses
import pprint
import json
import os
import subprocess
import argparse
import sys
import typing
import re
from pathlib import Path

from autopilot.setup.forms import Autopilot_Setup, DIRECTORY_STRUCTURE
from autopilot.setup.run_script import run_script, list_scripts, run_scripts
from autopilot.prefs import _DEFAULTS, Scopes


def make_dir(adir:Path, permissions:int=0o777):
    """
    Make a directory if it doesn't exist and set its permissions to `0777`

    Args:
        adir (str): Path to the directory
        permissions (int): an octal integer used to set directory permissions (default ``0o777``)
    """
    adir = Path(adir)
    if not adir.exists():
        adir.mkdir(parents=True,exist_ok=True)

    adir.chmod(permissions)

def make_alias(launch_script: Path, bash_profile: typing.Optional[str]=None) -> typing.Tuple[bool, str]:
    """
    Make an alias so that calling ``autopilot`` calls ``autopilot_dir/launch_autopilot.sh``

    Arguments:
        launch_script (str): the path to the autopilot launch script to be aliased
        bash_profile (str, None): Optional, location of shell profile to edit. if None, use ``.bashrc`` then ``.bash_profile`` if they exist
    """
    result = True
    message = 'alias for autopilot successfully created, open autopilot by calling upon it by its name like an old friend ;)'

    # find bash file
    if bash_profile is None:
        if (Path.home() / '.bashrc').exists():
            bash_profile = Path.home() / '.bashrc'
        elif  (Path.home() / '.bash_profile').exists():
            bash_profile = Path.home() / '.bash_profile'
        else:
            result = False
            message = 'No bash_profile provided and cant find in default locations! couldnt make alias'
            return result, message


    with open(bash_profile, 'r') as pfile:
        profile = pfile.read()

    # remove any previously set autopilot alias
    re.sub('\n# autopilot alias generated by setup_autopilot.py.*\nalias autopilot.*', '', profile)

    # make and append alias to profile

    profile = profile + f"\n# autopilot alias generated by setup_autopilot.py on {datetime.datetime.now().isoformat()}\nalias autopilot={Path(launch_script).resolve()}\n"
    try:
        with open(bash_profile, 'w') as pfile:
            pfile.write(profile)
    except PermissionError as e:
        result = False
        message = f"Got a permission error trying to write into the .bashrc found at {bash_profile}"

    return result, message


def parse_manual_prefs(manual_prefs:typing.List[str]) -> dict:
    prefdict = {}
    for p in manual_prefs:
        key, val = p.split('=')
        prefdict[key] = val
    return prefdict





def parse_args():
    parser = argparse.ArgumentParser(description="Setup an Autopilot Agent")
    parser.add_argument('-f', '--prefs',
                        help="Location of .json prefs file (default: ~/autopilot/prefs.json")
    parser.add_argument('-d', '--dir',
                        help="Autopilot directory (default: ~/autopilot)")
    parser.add_argument('-p', '--pref',
                        help="Specify a pref manually like -p PREFNAME=value",
                        action='append')
    parser.add_argument('-s', '--scripts',
                        help="Run setup scripts without entering a full setup routine. specify one or more like -s script1 -s script2 for available scripts see -l",
                        action='append')
    parser.add_argument('-l', '--list_scripts',
                        help="list available setup scripts!",
                        action='store_true')
    parser.add_argument('-q', '--quiet',
                        help="run setup in quiet mode, dont launch the interactive prompt, just use the prefs specified by -p and -f, and the scripts specified by -s",
                        action='store_true')

    args = parser.parse_args()
    if args.pref:
        args.pref = parse_manual_prefs(args.pref)
    if args.list_scripts:
        list_scripts()
        sys.exit()
    elif args.scripts and not args.quiet:
        run_scripts(args.scripts)
        sys.exit()

    return args

def locate_user_dir(args) -> Path:

    if args.dir:
        autopilot_dir = Path(args.dir).resolve()

    else:
        # check for a ~/.autopilot file that should point us to the autopilot directory if it exists
        autopilot_conf_fn = Path().home() / '.autopilot'
        if autopilot_conf_fn.exists():
            with open(autopilot_conf_fn, 'r') as aconf:
                autopilot_dir = Path(aconf.read()).resolve()
                # autopilot_dir = autopilof_conf['AUTOPILOTDIR']
        else:
            autopilot_dir = Path().home() / 'autopilot/'

    make_dir(autopilot_dir)

    return autopilot_dir

def run_form(prefs:dict) -> typing.Tuple[dict, typing.List[str]]:
    ###################################3
    # Run the npyscreen prompt

    try:
        setup = Autopilot_Setup(prefs)
        setup.run()
    except (_curses.error, nps.wgwidget.NotEnoughSpaceForWidget) as e:
        # get minimum column count
        try:
            min_cols = setup.getForm(setup.STARTING_FORM).min_c
            print(f'Problem opening the Setup GUI!\nThis is most likely due to this window not being wide enough\n' + \
                  'The minimum width for the setup GUI is:\n\033[0;32;40m' + \
                  "-" * min_cols + '\u001b[0m\n\n' + f'Got error:\n{e}')

        except:
            print(
                f'Problem opening the Setup GUI!\nThis is most likely due to this window not being wide enough\n\nGot Error:\n{e}')

        sys.exit()


    ####################################
    # Collect values
    new_prefs = setup.unpack_prefs()
    active_scripts = setup.active_scripts
    return new_prefs, active_scripts


def make_launch_script(prefs:dict, prefs_fn=None, launch_file=None, permissions:int=0o775) -> Path:

    # Create a launch script
    if prefs_fn is None:
        prefs_fn = Path(prefs['BASEDIR']) / 'prefs.json'
    else:
        prefs_fn = Path(prefs_fn)

    if launch_file is None:
        launch_file = Path(prefs['BASEDIR']) / 'launch_autopilot.sh'
    else:
        launch_file = Path(launch_file)

    if prefs['AGENT'] in ('PILOT', 'CHILD'):
        with open(launch_file, 'w') as launch_file_open:
            launch_file_open.write('#!/bin/bash\n')
            launch_file_open.write('killall jackd\n')
            launch_file_open.write('sudo killall pigpiod\n')
            launch_file_open.write('sudo mount -o remount,size=128M /dev/shm\n')
            if prefs['VENV']:
                launch_file_open.write("source " + os.path.join(prefs['VENV'], 'bin', 'activate')+'\n')
            launch_file_open.write('python3 -m autopilot.core.pilot -f {}'.format(prefs_fn))

    elif prefs['AGENT'] == 'TERMINAL':
        with open(launch_file, 'w') as launch_file_open:
            launch_file_open.write('#!/bin/bash\n')
            if prefs['VENV']:
                launch_file_open.write("source " + os.path.join(prefs['VENV'], 'bin', 'activate')+'\n')
            launch_file_open.write("python3 -m autopilot.core.terminal -f " + str(prefs_fn) + "\n")

    os.chmod(launch_file, permissions)
    return launch_file

def make_systemd(prefs:dict, launch_file:Path) -> typing.Tuple[bool, str]:
    success = False
    message = ''

    systemd_string = '''
[Unit]
Description=autopilot
After=multi-user.target

[Service]
Type=idle
ExecStart={launch_pi}

Restart=on-failure

[Install]
WantedBy=multi-user.target'''.format(launch_pi=launch_file)

    unit_loc = '/lib/systemd/system/autopilot.service'

    try:

        subprocess.call('sudo sh -c \"echo \'{}\' > {}\"'.format(systemd_string, unit_loc), shell=True)
        # enable the service
        subprocess.call(['sudo', 'systemctl', 'daemon-reload'])
        sysd_result = subprocess.call(['sudo', 'systemctl', 'enable', 'autopilot.service'])

        if sysd_result != 0:
            message = 'Systemd service could not be enabled :('
        else:
            success = True
            message = 'Systemd service installed and enabled, unit file written to {}'.format(unit_loc)

    except PermissionError:
        message = "systemd service could not be installed due to a permissions error.\n" + \
                          "create a unit file containing the following at {}\n\n{}".format(unit_loc, systemd_string)

    return success, message

def results_string(env_results:dict,
                   config_msgs:typing.List[str],
                   error_msgs:typing.List[str],
                   prefs_fn:str,
                   prefs) -> str:

    env_result = '\n----------------------------------------\n'
    env_result += 'prefs.json has been created and saved to {}\n'.format(prefs_fn)
    env_result += pprint.pformat(prefs)
    env_result += '\n----------------------------------------\n'

    env_result += "\033[0;32;40m\n--------------------------------\nEnvironment Configuration:\n"
    for config, result in env_results.items():
        if result:
            env_result += "  [ SUCCESS ] "
        else:
            env_result += "  [ FAILURE ] "

        env_result += config
        env_result += '\n'

    if prefs.get('VENV', False):
        env_result +=f"  [ SUCCESS ] virtualenv detected, path: {prefs['VENV']}\n"
    else:
        env_result += "  [ CMONDOG ]\033[1;37;41m no virtualenv detected\u001b[0m, running autopilot outside a venv is not recommended but it might work who knows\n"

    if len(config_msgs)>0:
        env_result += '\nAdditional Messages:'
        for msg in config_msgs:
            env_result += '  '
            env_result += msg
            env_result += '\n'

    if len(error_msgs)>0:
        env_result += '\n--------------------------------\n'
        for i, msg in enumerate(error_msgs):
            env_result += '\033[1;37;41mSomething went wrong during setup, this is wrong thing #{}\u001b[0m'.format(i)
            env_result += '\033[0;31;40m\n{}\n\u001b[0m'.format(msg)

    env_result += "\u001b[0m"

    return env_result

def make_ectopic_dirnames(basedir:Path) -> dict:
    basedir = Path(basedir).resolve()
    out = {}
    out['BASEDIR'] = str(basedir)
    for key, val in _DEFAULTS.items():
        if val['scope'] == Scopes.DIRECTORY and key != "BASEDIR":
            out[key] = str(basedir / Path(val['default']).stem)
    return out


def main():
    env = {}
    env_results = {}
    prefs = {}
    extra_dirs = {}
    error_msgs = []
    config_msgs = []

    args = parse_args()

    autopilot_dir = locate_user_dir(args)
    # if we were passed an explicit basedir, make a dict of the usual directory structure
    # to use as defaults
    if args.dir is not None:
        extra_dirs = make_ectopic_dirnames(autopilot_dir)

    # attempt to load .prefs from standard location (~/autopilot/prefs.json)
    if args.prefs:
        prefs_fn = Path(args.prefs).resolve()
    else:
        prefs_fn = autopilot_dir / 'prefs.json'

    if prefs_fn.exists():
        if not args.quiet:
            print(f'Existing prefs found, loading from {prefs_fn}')
        with open(prefs_fn, 'r') as prefs_f:
            prefs = json.load(prefs_f)
    else:
        if not args.quiet:
            print('No existing prefs found, starting from defaults')

    # recombine default directories after loading prefs, but before
    # incorporating any custom passed prefs.
    prefs.update(extra_dirs)

    # combine and manually given prefs
    if args.pref is not None:
        prefs.update(args.pref)

    #################
    # run interactive form unless we were passed one or both of the things it gives us (prefs and scripts)

    if not args.quiet:
        new_prefs, active_scripts = run_form(prefs)
        prefs.update(new_prefs)
    else:
        if args.scripts is not None:
            active_scripts = args.scripts
        else:
            active_scripts = []


    ####################################
    # Configure Environment

    # run our scripts
    env_results.update(run_scripts(active_scripts, return_all=True, print_status=False))

    # Create directory structure if needed
    dirs_to_make = [path for dir_name, path in prefs.items() if dir_name in DIRECTORY_STRUCTURE.keys()]
    print('Creating Directories: \n' + '\n'.join(dirs_to_make))

    for make_this_dir in dirs_to_make:
        make_dir(make_this_dir)

    # make our launch script
    launch_file = make_launch_script(prefs)
    config_msgs.append("Launch file created at {}".format(launch_file))

    # install as systemd service if requested
    if 'systemd' in active_scripts:
        made_systemd, msg = make_systemd(prefs, launch_file)
        env_results['systemd'] = made_systemd
        if made_systemd:
            config_msgs.append(msg)
        else:
            error_msgs.append(msg)

    if 'alias' in active_scripts:
        # make an alias for autopilot!
        made_alias, msg = make_alias(launch_file)
        env_results['alias'] = made_alias
        if made_alias:
            config_msgs.append(msg)
        else:
            error_msgs.append(msg)

    ####################################
    # save prefs and finalize environment

    # save prefs
    with open(prefs_fn, 'w') as prefs_f:
        json.dump(prefs, prefs_f, indent=4, separators=(',', ': '), sort_keys=True)

    # save basedir in autopilot user file
    with open(os.path.join(os.path.expanduser('~'), '.autopilot'), 'w') as autopilot_f:
        autopilot_f.write(prefs['BASEDIR'])

    #####################################3
    # User feedback
    result_str = results_string(env_results, config_msgs, error_msgs, prefs_fn, prefs)
    print(result_str)


if __name__ == "__main__":
    main()




import typing
from time import time

import numpy as np
from scipy.spatial import distance
from scipy.spatial.transform import Rotation as R
from scipy.optimize import curve_fit

from autopilot.transform.transforms import Transform
from autopilot.transform.timeseries import Kalman


class Distance(Transform):
    """
    Given an n_samples x n_dimensions array, compute pairwise or mean distances
    """
    format_in = {'type': np.ndarray}
    format_out = {'type': np.ndarray}

    def __init__(self,
                 pairwise: bool=False,
                 n_dim: int = 2,
                 metric: str='euclidean',
                 squareform: bool=True,
                 *args, **kwargs):
        """

        Args:
            pairwise (bool): If False (default), return mean distance. if True, return all distances
            n_dim (int): number of dimensions (input array will be filtered like ``input[:,0:n_dim]``
            metric (str): any metric acceptable to :func:`scipy.spatial.distance.pdist
            squareform (bool): if pairwise is True, if True return square distance matrix, otherwise return compressed distance matrix (dist(X[i], X[j] = y[i*j])
            *args:
            **kwargs:
        """

        super(Distance, self).__init__(*args, **kwargs)

        self.pairwise = pairwise
        self.n_dim = n_dim
        self.metric = metric
        self.squareform = squareform

    def process(self, input: np.ndarray):

        # filter to input_dimension
        input = input[:,0:self.n_dim]

        output = distance.pdist(input, metric=self.metric)

        if self.pairwise:
            if self.squareform:
                output = distance.squareform(output)
        else:
            output = np.mean(output)

        return output


class Angle(Transform):
    """
    Get angle between line formed by two points and horizontal axis
    """

    format_in = {'type': np.ndarray}
    format_out = {'type': float}

    def __init__(self, abs=True, degrees=True, *args, **kwargs):
        super(Angle, self).__init__(*args, **kwargs)
        self.abs = abs
        self.degrees = degrees

    def process(self, input):
        angle = np.arctan2(input[1][1]-input[0][1], input[1][0]-input[0][0])
        if self.abs:
            angle += np.pi
        if self.degrees:
            angle = angle*(180/np.pi)
        return angle


class IMU_Orientation(Transform):
    """
    Compute absolute orientation (roll, pitch) from accelerometer and gyroscope measurements
    (eg from :class:`.hardware.i2c.I2C_9DOF` )

    Uses a :class:`.timeseries.Kalman` filter, and implements :cite:`patonisFusionMethodCombining2018a` to fuse
    the sensors

    Can be used with accelerometer data only, or with combined accelerometer/gyroscope data for
    greater accuracy

    Arguments:
        invert_gyro (bool): if the gyroscope's orientation is inverted from accelerometer measurement, multiply
            gyro readings by -1 before using
        use_kalman (bool): Whether to use kalman filtering (True, default), or return raw trigonometric
            transformation of accelerometer readings (if provided, gyroscope readings will be ignored)

    Attributes:
        kalman (:class:`.transform.timeseries.Kalman`): If ``use_kalman == True`` , the Kalman Filter.

    References:
        :cite:`patonisFusionMethodCombining2018a`
        :cite:`abyarjooImplementingSensorFusion2015`
    """

    def __init__(self, use_kalman:bool = True, invert_gyro:bool=False, *args, **kwargs):
        super(IMU_Orientation, self).__init__(*args, **kwargs)

        self.invert_gyro = invert_gyro # type: bool
        self._last_update = None # type: typing.Optional[float]
        self._dt = 0 # type: float
        # preallocate orientation array for filtered values
        self.orientation = np.zeros((2), dtype=float) # type: np.ndarray
        # and for unfiltered values so they aren't ambiguous
        self._orientation = np.zeros((2), dtype=float)  # type: np.ndarray

        self.kalman = None # type: typing.Optional[Kalman]
        if use_kalman:
            self.kalman = Kalman(dim_state=2, dim_measurement=2, dim_control=2)  # type: typing.Optional[Kalman]

    def process(self, accelgyro:typing.Union[typing.Tuple[np.ndarray, np.ndarray], np.ndarray]) -> np.ndarray:
        """

        Args:
            accelgyro (tuple, :class:`numpy.ndarray`): tuple of (accelerometer[x,y,z], gyro[x,y,z]) readings as arrays, or
                an array of just accelerometer[x,y,z]

        Returns:
            :class:`numpy.ndarray`: filtered [roll, pitch] calculations in degrees
        """
        # check what we were given...
        if isinstance(accelgyro, (tuple, list)) and len(accelgyro) == 2:
            # combined accelerometer and gyroscope readings
            accel, gyro = accelgyro
        elif isinstance(accelgyro, np.ndarray) and np.squeeze(accelgyro).shape[0] == 3:
            # just accelerometer readings
            accel = accelgyro
            gyro = None
        else:
            # idk lol
            self.logger.exception(f'Need input to be a tuple of accelerometer and gyroscope readings, or an array of accelerometer readings. got {accelgyro}')
            return

        # convert accelerometer readings to roll and pitch
        pitch = 180*np.arctan2(accel[0], np.sqrt(accel[1]**2 + accel[2]**2))/np.pi
        roll = 180*np.arctan2(accel[1], np.sqrt(accel[0]**2 + accel[2]**2))/np.pi


        if self.kalman is None:
            # store orientations in external attribute if not using kalman filter
            self.orientation[:] = (roll, pitch)
            return self.orientation.copy()
        else:
            # if using kalman filter, use private array to store raw orientation
            self._orientation[:] = (roll, pitch)

        # TODO: Don't assume that we're fed samples instantatneously -- ie. once data representations are stable, need to accept a timestamp here rather than making one
        if self._last_update is None or gyro is None:
            # first time through don't have dt to scale gyro by
            self.orientation[:] = np.squeeze(self.kalman.process(self._orientation))
            self._last_update = time()

        else:
            if self.invert_gyro:
                gyro *= -1

            # get dt for time since last update
            update_time = time()
            self._dt = update_time-self._last_update
            self._last_update = update_time

            if self._dt>1:
                # if it's been really long, the gyro read is pretty much useless and will give ridiculous reads
                self.orientation[:] = np.squeeze(self.kalman.process(self._orientation))
            else:
                # run predict and update stages separately to incorporate gyro
                self.kalman.predict(u=gyro[0:2]*self._dt)
                self.orientation[:] = np.squeeze(self.kalman.update(self._orientation))

        return self.orientation.copy()


class Rotate(Transform):
    """
    Rotate in 3 dimensions using :class:`scipy.spatial.transform.Rotation`

    Args:
        dims ( "xyz" ): string specifying which axes the rotation will be around, eg ``"xy"`` , ``"xyz"```
        rotation_type (str): Format of rotation input, must be one available to the :class:`~scipy.spatial.transform.Rotation` class
            (but currently only euler angles are supported)
        degrees (bool): whether to output rotation in degrees (True, default) or radians
        inverse ("xyz"): dimensions in the "rotation" input to :meth:`.Rotate.process` to inverse before applying rotation
        rotation (tuple, list, :class:`numpy.ndarray`, None): If supplied, use the same rotation for all processed data. If None,
            :meth:`.Rotate.process` will expect a tuple of (data, rotation).
    """

    _DIMS = {
        'x': 0,
        'y': 1,
        'z': 2
    }

    def __init__(self, dims="xyz", rotation_type="euler", degrees=True, inverse="", rotation=None, *args, **kwargs):
        super(Rotate, self).__init__(*args, **kwargs)

        self.degrees = degrees
        self.rotation_type = rotation_type

        # parse dimensions and inverse into slices
        if not dims:
            e = ValueError('need to provide some dimensino to rotate around, got empty dims')
            self.logger.exception(e)
            raise e

        # store dims and something we can slice with for dims and inverse
        self.dims = dims
        self._dims = [self._DIMS[dim] for dim in dims]

        if not inverse:
            self.inverse = False
            self._inverse = None
        else:
            self.inverse = inverse
            self._inverse = [self._DIMS[dim] for dim in inverse]

        # stash rotation creation method depending on rotation_type
        if rotation_type == "euler":
            self._rotate_constructor = R.from_euler
        else:
            e = NotImplementedError('Only euler is implemented currently!')
            self.logger.exception(e)
            raise e

        # if we were provided an initial rotation, instantiate rotation here
        if rotation:
            # inverse what must be inverted
            if self.inverse:
                rotation[self._inverse] *= -1
            self._rotation = rotation
            self._rotator = self._rotate_constructor(self.dims, self._rotation, degrees=self.degrees)
        else:
            self._rotation = None
            self._rotator = None

    def process(self, input):
        """


        Args:
            input (tuple, :class:`numpy.ndarray`): a tuple of (input[x,y,z], rotation[x,y,z]) where input is to be rotated
                according to the axes in rotation (indicated in :attr:`.Rotate.dims` ). If only an input array is provided,
                a static rotation array must have been provided in the constructor (otherwise the most recent rotation will be used)

        Returns:
            :class:`numpy.ndarray` - rotated input array
        """

        if isinstance(input, (tuple, list)) and len(input) == 2:
            # split out input coords and rotation
            input, rotate = input

            # invert what must be inverted
            if self.inverse:
                rotate[self._inverse] *= -1

        else:
            rotate = None

        # if given a new rotation, use it
        if rotate is not None and (self._rotation is None or not np.array_equal(rotate, self._rotation)):
            self._rotator = self._rotate_constructor(self.dims, rotate, degrees=self.degrees)
            self._rotation = rotate

        # apply itttt and return
        try:
            return self._rotator.apply(input)
        except AttributeError:
            if self._rotator is None:
                e = RuntimeError('No rotation was provided, and none is available!')
                self.logger.exception(e)
                raise e

class Spheroid(Transform):
    """
    Fit and transform 3d coordinates according to some spheroid.

    Eg. for calibrating accelerometer readings by transforming them from their uncalibrated spheroid to the expected
    sphere with radius == 9.8m/s/s centered at (0,0,0).

    Does not estimate/correct for rotation of the spheroid.

    Examples:
        ######
        # Calibrate an accelerometer by transforming readings to a 9.8-radius sphere centered at 0
        >>> sphere = Spheroid(target=(9.8,9.8,9.8,0,0,0))
        # take some readings...
        # imagine we're taking them from some sensor idk
        # say our sensor slightly exaggerates gravity in the z-axis...
        >>> readings = np.array((0.,0.,10.5))
        # fit our object (need >>1 sample)
        >>> sphere.fit(readings)
        # transform to proper gravity
        >>> sphere.process(readings)
        [0., 0., 9.8]

    Args:
        target (tuple): parameterization of spheroid to transform to, if none is passed, transform to unit circle
            centered at (0,0,0). parameterized as::

            (a, # radius of x dimension
            b, # radius of y dimension
            c, # radius of z dimension
            x, # x-offset
            y, # y-offset
            z) # z-offset

        source (tuple): parameterization of spheroid to transform from in the same 6-tuple form as ``target``,
            if None is passed, assume we will use :meth:`.Spheroid.fit`
        fit (None, :class:`numpy.ndarray`): Initialize with values to fit, if None assume fit will be called later.


    References:
        * https://jekel.me/2020/Least-Squares-Ellipsoid-Fit/
        * http://www.juddzone.com/ALGORITHMS/least_squares_3D_ellipsoid.html
    """

    def __init__(self, target=(1,1,1,0,0,0),
                 source:tuple=(None, None, None, None, None, None),
                 fit:typing.Optional[np.ndarray]=None,
                 *args, **kwargs):
        super(Spheroid, self).__init__(*args, **kwargs)

        self.target = target
        self.source  = source

        self._scale = None
        self._offset_source = None
        self._offset_target = None
        self._update_arrays()

        if fit is not None:
            self.fit(fit, **kwargs)

    def _update_arrays(self):
        if not any([val is None for val in self.source]):
            self._scale = np.array((self.target[0]/self.source[0],
                                    self.target[1]/self.source[1],
                                    self.target[2]/self.source[2]))
            self._offset_source = np.array((self.source[3], self.source[4], self.source[5]))
            self._offset_target = np.array((self.target[3], self.target[4], self.target[5]))

    def fit(self, points, **kwargs):
        """
        Fit a spheroid from a set of noisy measurements

        updates the :attr:`._scale` and :attr:`._offset` private arrays used to manipulate input data

        .. note::

            It's usually important to pass ``bounds`` to :func:`scipy.optimize.curve_fit` !!! passed as a 2-tuple
            of ``((min_a, min_b, ...), (max_a, max_b...))`` In particular such that a, b, and c are positive. If no
            bounds are passed, assume at least that much.

        Args:
            points (:class:`numpy.ndarray`): (M, 3) array of points to fit
            **kwargs (): passed on to :func:`scipy.optimize.curve_fit`

        Returns:
            tuple: parameters of fit ellipsoid (a,b,c,x,y,z)
        """
        if 'bounds' in kwargs.keys():
            bounds = kwargs.pop('bounds')
        else:
            bounds = ((0,      0,      0,      -np.inf, -np.inf, -np.inf),
                      (np.inf, np.inf, np.inf,  np.inf,  np.inf,  np.inf))

        y = np.ones((points.shape[0]))
        parameters, _ = curve_fit(_ellipsoid_func, points, y, bounds=bounds, **kwargs)
        self.source = parameters
        self._update_arrays()

    def process(self, input:np.ndarray):
        """
        Transform input (x,y,z) points such that points in :attr:`.source` are mapped to those in :attr:`.target`

        Args:
            input (:class:`numpy.ndarray`): x, y, and z coordinates

        Returns:
            :class:`numpy.ndarray` : coordinates transformed according to the spheroid requested
        """
        if self._scale is None or self._offset_target is None or self._offset_source is None:
            self.logger.exception('process called without fit being performed or source ellipsoid provided! returning untransformed points!')
            return input

        # move to the center, then scale, then offset.

        return ((input - self._offset_source) * self._scale) + self._offset_target

    def generate(self, n:int, which:str='source', noise:float=0):
        """
        Generate random points from the ellipsoid

        Args:
            n (int): number of points to generate
            which ('str'): which spheroid to generate from? ('source' - default, or 'target')
            noise (float): noise to add to points

        Returns:
            :class:`numpy.ndarray` : (n, 3) array of generated points
        """
        if which == "source":
            if not any([val is None for val in self.source]):
                a,b,c,x,y,z = self.source
            else:
                self.logger.exception('Cannot generate from source, dont have ellipsoid parameterization')
                return
        elif which == "target":
            a, b, c, x, y, z = self.target
        else:
            self.logger.exception(f"Dont know how to generate points for which == {which}")
            return

        u = np.random.rand(n)
        v = np.random.rand(n)
        theta = u * 2.0 * np.pi
        phi = np.arccos(2.0 * v - 1.0)
        sinTheta, cosTheta = np.sin(theta), np.cos(theta)
        sinPhi,   cosPhi   = np.sin(phi),   np.cos(phi)
        rx = (a * sinPhi * cosTheta) + x + (np.random.rand(n) * noise)
        ry = (b * sinPhi * sinTheta) + y + (np.random.rand(n) * noise)
        rz = (c * cosPhi) + z + (np.random.rand(n) * noise)
        return np.column_stack((rx, ry, rz))



def _ellipsoid_func(fit, a, b, c, x, y, z):
    """
    Ellipsoid equation for use with :meth:`.Ellipsoid.fit`

    Args:
        fit (:class:`numpy.ndarray`): (M, 3) array of x,y,z points to fit
        a (float): X-scale parameter to fit
        b (float): Y-scale parameter to fit
        c (float): Z-scale parameter to fit
        x (float): X-offset parameter to fit
        y (float): Y-offset parameter to fit
        z (float): Z-offset parameter to fit

    Returns:
        float: result of ellipsoid function, minimize parameters to == 1
    """
    x_fit, y_fit, z_fit = fit[:,0], fit[:,1], fit[:,2]
    return ((x_fit - x)**2 / a**2) + ((y_fit - y)**2 / b**2) + ((z_fit - z)**2 / c**2)
















"""This module defines classes to generate different sounds.

These classes are currently implemented:
* Tone : a sinuosoidal pure tone
* Noise : a burst of white noise
* File : read from a file
* Speech
* Gap

The behavior of this module depends on `prefs.get('AUDIOSERVER')`.
* If this is 'jack', or True:
    Then import jack, define Jack_Sound, and all sounds inherit from that.
* If this is 'pyo':
    Then import pyo, define PyoSound, and all sounds inherit from that.
* If this is 'docs':
    Then import both jack and pyo, define both Jack_Sound and PyoSound,
    and all sounds inherit from `object`.
* Otherwise:
    Then do not import jack or pyo, or define either Jack_Sound or PyoSound,
    and all sounds inherit from `object`.

TODO:
    Implement sound level and filter calibration
"""

# Re: The organization of this module
# We balance a few things:
# 1) using two sound servers with very different approaches to
# delivering sounds, and
# 2) having a similar API so other modules can query sound properties
# while still being agnostic to the sound server.
# 3) not have our classes split into a ton of Pyo_Tone, Jack_Tone
# copies so they have their parameters and behavior drift apart
#
# So, We have base classes, but they can't encapsulate all the
# behavior for making sounds, so use an init_audio() method that
# creates sound conditional on the type of audio server.


import os
import sys
import typing
from time import sleep
from scipy.io import wavfile
from scipy.signal import resample
import numpy as np
import threading
from itertools import cycle
from queue import Empty, Full

from autopilot import prefs
from autopilot.core.loggers import init_logger
from autopilot.stim.stim import Stim
import autopilot

## First, switch the behavior based on the pref AUDIOSERVER
# Get the pref
server_type = prefs.get('AUDIOSERVER')

# True is a synonym for 'jack', the default server
if server_type == True:
    server_type = 'jack'

# Force lower-case if string
try:
    server_type = server_type.lower()
except AttributeError:
    pass

# From now on, server_type should be 'jack', 'pyo', 'docs', or None
if server_type not in ['jack', 'pyo', 'docs']:
    server_type = None

# if we're testing, set server_type to jack
if 'pytest' in sys.modules:
    server_type = 'jack'


## Import the required modules
if server_type in ['jack', 'docs']:
    # This will warn if the jack library is not found
    from autopilot.stim.sound import jackclient

elif server_type in ['pyo', 'docs']:
    # Using these import guards for compatibility, but I think we should
    # actually error here
    try:
        import pyo
    except ImportError:
        pass


## Define Pyo_Sound if needed
if server_type in ("pyo", "docs"):
    class Pyo_Sound(object):
        """
        Metaclass for pyo sound objects.

        Note:
            Use of pyo is generally discouraged due to dropout issues and
            the general opacity of the module. As such this object is
            intentionally left undocumented.

        """
        def __init__(self):
            self.PARAMS = None  # list of strings of parameters to be defined
            self.type = None  # string human readable name of sound
            self.duration = None  # duration in ms
            self.amplitude = None
            self.table = None
            self.trigger = None
            self.server_type = 'pyo'

        def play(self):
            self.table.out()

        def table_wrap(self, audio, duration=None):
            """Records a PyoAudio generator into a sound table, returns a
            tableread object which can play the audio with .out()

            Args:
                audio:
                duration:
            """

            if not duration:
                duration = self.duration

            # Duration is in ms, so divide by 1000
            # See https://groups.google.com/forum/#!topic/pyo-discuss/N-pan7wPF-o
            # TODO: Get chnls to be responsive to NCHANNELS in prefs. hardcoded for now
            tab = pyo.NewTable(length=(float(duration) / 1000),
                               chnls=prefs.get('NCHANNELS'))  # Prefs should always be declared in the global namespace
            tabrec = pyo.TableRec(audio, table=tab, fadetime=0.005).play()
            sleep((float(duration) / 1000))
            self.table = pyo.TableRead(tab, freq=tab.getRate(), loop=0)

        def set_trigger(self, trig_fn):
            """
            Args:
                trig_fn:
            """
            # Using table triggers, call trig_fn when table finishes playing
            self.trigger = pyo.TrigFunc(self.table['trig'], trig_fn)


## Define Jack_Sound if needed
if server_type in ("jack", "docs"):
    class Jack_Sound(object):
        """
        Base class for sounds that use the :class:`~.jackclient.JackClient` audio
        server.

        Attributes:
            PARAMS (list): List of strings of parameters that need to be defined for this sound
            type (str): Human readable name of sound type
            duration (float): Duration of sound in ms
            amplitude (float): Amplitude of sound as proportion of 1 (eg 0.5 is half amplitude)
            table (:class:`numpy.ndarray`): A Numpy array of samples
            chunks (list): :attr:`~.Jack_Sound.table` split up into chunks of :data:`~.jackclient.BLOCKSIZE`
            trigger (callable): A function that is called when the sound completes
            nsamples (int): Number of samples in the sound
            padded (bool): Whether the sound had to be padded with zeros when split into chunks (ie. sound duration was not a multiple of BLOCKSIZE).
            fs (int): sampling rate of client from :data:`.jackclient.FS`
            blocksize (int): blocksize of client from :data:`.jackclient.BLOCKSIZE`
            server (:class:`~.jackclient.Jack_Client`): Current Jack Client
            q (:class:`multiprocessing.Queue`): Audio Buffer queue from :data:`.jackclient.QUEUE`
            q_lock (:class:`multiprocessing.Lock`): Audio Buffer lock from :data:`.jackclient.Q_LOCK`
            play_evt (:class:`multiprocessing.Event`): play event from :data:`.jackclient.PLAY`
            stop_evt (:class:`multiprocessing.Event`): stop event from :data:`.jackclient.STOP`
            buffered (bool): has this sound been dumped into the :attr:`~.Jack_Sound.q` ?
            buffered_continuous (bool): Has the sound been dumped into the :attr:`~.Jack_Sound.continuous_q`?

        """

        PARAMS = []
        """
        list:  list of strings of parameters to be defined
        """

        type = None
        """
        str: string human readable name of sound
        """

        server_type = 'jack'
        """
        str: type of server, always 'jack' for `Jack_Sound` s.
        """

        def __init__(self):
            """Initialize a new Jack_Sound
            
            This sets sound-specific parameters to None, set jack-specific
            parameters to their equivalents in jackclient, initializes
            some other flags and a logger.
            """
            # These sound-specific parameters will be set by the derived
            # objects.
            self.duration = None  # duration in ms
            self.amplitude = None
            self.table = None  # numpy array of samples
            self.chunks = None  # table split into a list of chunks
            self.trigger = None
            self.nsamples = None
            self.padded = False # whether or not the sound was padded with zeros when chunked
            self.continuous = False

            # These jack-specific parameters are copied from jackclient
            self.fs = jackclient.FS
            self.blocksize = jackclient.BLOCKSIZE
            self.server = jackclient.SERVER
            self.q = jackclient.QUEUE
            self.q_lock = jackclient.Q_LOCK
            self.play_evt = jackclient.PLAY
            self.stop_evt = jackclient.STOP
            self.continuous_flag = jackclient.CONTINUOUS
            self.continuous_q = jackclient.CONTINUOUS_QUEUE
            self.continuous_loop = jackclient.CONTINUOUS_LOOP

            # Initalize these flags
            self.initialized = False
            self.buffered = False
            self.buffered_continuous = False

            # Initialize a logger
            self.logger = init_logger(self)

        def chunk(self, pad=True):
            """
            Split our `table` up into a list of :attr:`.Jack_Sound.blocksize` chunks.

            Args:
                pad (bool): If the sound is not evenly divisible into chunks, 
                pad with zeros (True, default), otherwise jackclient will pad 
                with its continuous sound
            """
            # Convert the table to float32 (if it isn't already)
            sound = self.table.astype(np.float32)

            # Determine how much longer it would have to be, if it were
            # padded to a length that is a multiple of self.blocksize
            oldlen = len(sound)
            newlen = int(
                np.ceil(float(oldlen) / self.blocksize) * self.blocksize)

            
            ## Only pad if necessary AND requested
            if pad and newlen > oldlen:
                # Pad differently depending on mono or stereo
                if sound.ndim == 1:
                    # Pad with 1d array of zeros
                    to_concat = np.zeros((newlen - oldlen,), np.float32)

                    # Pad
                    sound = np.concatenate([sound, to_concat])

                elif sound.ndim == 2:
                    # Each column is a channel
                    n_channels = sound.shape[1]
                    
                    # Raise error if more than two-channel sound
                    # This would actually be fine for this function, but this 
                    # almost surely indicates somebody has transposed something
                    if n_channels > 2:
                        raise ValueError("only 1- or 2-channel sound supported")
                    
                    # Pad with 2d array of zeros
                    to_concat = np.zeros(
                        (newlen - oldlen, sound.shape[1]), np.float32)

                    # Pad
                    sound = np.concatenate([sound, to_concat])
                
                else:
                    raise ValueError("sound must be 1d or 2d")
                
                # Flag as padded
                self.padded = True
            
            else:
                # Flag as not padded
                self.padded = False
            
            
            ## Reshape into chunks, each of length `self.blocksize`
            if sound.ndim == 1:
                sound_list = [sound[i:i+self.blocksize] for i in range(0, sound.shape[0], self.blocksize)]
            
            elif sound.ndim == 2:
                sound_list = [sound[i:i + self.blocksize,:] for i in range(0, sound.shape[0], self.blocksize)]

            else:
                raise NotImplementedError(f"sounds with more than 2 dimensions arent supported, got ndim {sound.ndim}")

            self.chunks = sound_list

        def set_trigger(self, trig_fn):
            """
            Set a trigger function to be called when the :attr:`~.Jack_Sound.stop_evt` is set.

            Args:
                trig_fn (callable): Some callable
            """
            if callable(trig_fn):
                self.trigger = trig_fn
            else:
                Exception('trigger must be callable')

        def wait_trigger(self):
            """
            Wait for the stop_evt trigger to be set for at least a second after
            the sound should have ended.

            Call the trigger when the event is set.
            """
            # wait for our duration plus a second at most.
            self.stop_evt.wait((self.duration+1000)/1000.)
            # if the sound actually stopped...
            if self.stop_evt.is_set():
                self.trigger()

        def get_nsamples(self):
            """
            given our fs and duration, how many samples do we need?

            literally::

                np.ceil((self.duration/1000.)*self.fs).astype(np.int)

            """
            self.nsamples = np.ceil((self.duration/1000.)*self.fs).astype(np.int)

        def quantize_duration(self, ceiling=True):
            """
            Extend or shorten a sound so that it is a multiple of :data:`.jackclient.BLOCKSIZE`

            Args:
                ceiling (bool): If true, extend duration, otherwise decrease duration.
            """

            # get remainder of samples
            self.get_nsamples()
            remainder = self.nsamples % self.blocksize

            if remainder == 0:
                return

            # get target number of samples
            # get target n blocks and multiply by blocksize
            if ceiling:
                target_samples = np.ceil(float(self.nsamples)/self.blocksize)*self.blocksize
            else:
                target_samples = np.floor(float(self.nsamples)/self.blocksize)*self.blocksize

            # get new duration
            self.duration = (target_samples/self.fs)*1000.

            # refresh nsamples
            self.get_nsamples()

        def buffer(self):
            """
            Dump chunks into the sound queue.
            
            After the last chunk, a `None` is put into the queue. This
            tells the jack server that the sound is over and that it should
            clear the play flag.
            """

            if hasattr(self, 'path'):
                self.logger.debug('BUFFERING SOUND {}'.format(self.path))

            if not self.initialized and not self.table:
                try:
                    self.init_sound()
                    self.initialized = True
                except:
                    pass
                    #TODO: Log this, better error handling here

            if not self.chunks:
                self.chunk()

            with self.q_lock:
                # empty queue
                # FIXME: Testing whether this is where we get held up on the 'fail after sound play' bug
                # n_gets = 0
                while not self.q.empty():
                    try:
                        _ = self.q.get_nowait()
                    except Empty:
                        # normal, get until it's empty
                        break
                    # n_gets += 1
                    # if n_gets > 100000:
                    #     break
                for frame in self.chunks:
                    self.q.put_nowait(frame)
                # The jack server looks for a None object to clear the play flag
                self.q.put_nowait(None)
                self.buffered = True

        def buffer_continuous(self):
            """
            Dump chunks into the continuous sound queue for looping.

            Continuous shoulds should always have full frames -
            ie. the number of samples in a sound should be a multiple of :data:`.jackclient.BLOCKSIZE`.

            This method will call :meth:`.quantize_duration` to force duration such that the sound has full frames.

            An exception will be raised if the sound has been padded.
            """

            # FIXME: Initialized should be more flexible,
            # for now just deleting whatever init happened because
            # continuous sounds are in development
            self.table = None
            self.initialized = False

            if not self.initialized and not self.table:
                self.quantize_duration()
                self.init_sound()
                self.initialized = True

            if not self.chunks:
                self.chunk()

            # continous sounds should not have any padding - see docstring
            if self.padded:
                raise Exception("Continuous sounds cannot have padded chunks - sounds need to have n_samples % blocksize == 0")

            # empty queue
            while not self.continuous_q.empty():
                try:
                    _ = self.continuous_q.get_nowait()
                except Empty:
                    # normal, get until it's empty
                    break

            # put all the chunks into the queue, rather than one at a time
            # to avoid partial receipt
            self.continuous_q.put(self.chunks.copy())

            self.buffered_continuous = True

        def play(self):
            """
            Play ourselves.

            If we're not buffered, be buffered.

            Otherwise, set the play event and clear the stop event.

            If we have a trigger, set a Thread to wait on it.
            """
            if not self.buffered:
                self.buffer()

            if hasattr(self, 'path'):
                self.logger.debug('PLAYING SOUND {}'.format(self.path))

            self.play_evt.set()
            self.stop_evt.clear()
            self.buffered = False

            if callable(self.trigger):
                threading.Thread(target=self.wait_trigger).start()

        def play_continuous(self, loop=True):
            """
            Play the sound continuously.

            Sound will be paused if another sound has its 'play' method called.

            Currently - only looping is implemented: the full sound is loaded by the jack client and repeated indefinitely.

            In the future, sound generation methods will be refactored as python generators so sounds can be continuously generated and played.

            Args:
                loop (bool): whether the sound will be stored by the jack client and looped (True), or whether the sound will be continuously streamed (False, not implemented)

            Returns:

            todo::

                merge into single play method that changes behavior if continuous or not

            """

            if not loop:
                raise NotImplementedError('Continuous, unlooped streaming has not been implemented yet!')

            if not self.buffered_continuous:
                self.buffer_continuous()

            if loop:
                self.continuous_loop.set()
            else:
                self.continuous_loop.clear()

            # tell the sound server that it has a continuous sound now
            self.continuous_flag.set()
            self.continuous = True

            # after the sound server start playing, it will clear the queue, unbuffering us
            self.buffered_continuous = False

        def stop_continuous(self):
            """
            Stop playing a continuous sound

            Should be merged into a general stop method
            """
            if not self.continuous:
                self.logger.warning("stop_continuous called but not a continuous sound!")
                return

            self.continuous_flag.clear()
            self.continuous_loop.clear()

        def end(self):
            """
            Release any resources held by this sound

            """

            if self.play_evt.is_set():
                self.play_evt.clear()

            if not self.stop_evt.is_set():
                self.stop_evt.set()

            if self.continuous:
                while not self.continuous_q.empty():
                    try:
                        _ = self.continuous_q.get_nowait()
                    except Empty:
                        # normal, get until it's empty
                        break
                self.buffered_continuous = False
                self.continuous_flag.clear()

            self.table = None
            self.initialized = False

        def __del__(self):
            self.end()


## Now define BASE_CLASS to be Pyo_Sound, Jack_Sound, or object
if server_type == "pyo":
    BASE_CLASS = Pyo_Sound
elif server_type == "jack":
    BASE_CLASS = Jack_Sound
else:
    # just importing to query parameters, not play sounds.
    BASE_CLASS = Stim


## The rest of the module defines actual sounds, which inherit from BASE_CLASS
class Tone(BASE_CLASS):
    """The Humble Sine Wave"""

    PARAMS = ['frequency','duration','amplitude']
    type = 'Tone'

    def __init__(self, frequency, duration, amplitude=0.01, **kwargs):
        """
        Args:
            frequency (float): frequency of sin in Hz
            duration (float): duration of the sin in ms
            amplitude (float): amplitude of the sound as a proportion of 1.
            **kwargs: extraneous parameters that might come along with instantiating us
        """
        super(Tone, self).__init__()

        self.frequency = float(frequency)
        self.duration = float(duration)
        self.amplitude = float(amplitude)

        self.init_sound()

    def init_sound(self):
        """
        Create a sine wave table using pyo or numpy, depending on the server type.
        """

        if self.server_type == 'pyo':
            sin = pyo.Sine(self.frequency, mul=self.amplitude)
            self.table = self.table_wrap(sin)
        elif self.server_type == 'jack':
            self.get_nsamples()
            t = np.arange(self.nsamples)
            self.table = (self.amplitude*np.sin(2*np.pi*self.frequency*t/self.fs)).astype(np.float32)
            #self.table = np.column_stack((self.table, self.table))
            self.chunk()

        self.initialized = True

class Noise(BASE_CLASS):
    """Generates a white noise burst with specified parameters
    
    The `type` attribute is always "Noise".
    """
    # These are the parameters of the sound, I think this is used to generate
    # sounds automatically for a protocol
    PARAMS = ['duration','amplitude', 'channel']
    
    # The type of the sound
    type='Noise'
    
    def __init__(self, duration, amplitude=0.01, channel=None, **kwargs):
        """Initialize a new white noise burst with specified parameters.
        
        The sound itself is stored as the attribute `self.table`. This can
        be 1-dimensional or 2-dimensional, depending on `channel`. If it is
        2-dimensional, then each channel is a column.
        
        Args:
            duration (float): duration of the noise
            amplitude (float): amplitude of the sound as a proportion of 1.
            channel (int or None): which channel should be used
                If 0, play noise from the first channel
                If 1, play noise from the second channel
                If None, send the same information to all channels ("mono")
            **kwargs: extraneous parameters that might come along with instantiating us
        """
        # This calls the base class, which sets server-specific parameters
        # like samplign rate
        super(Noise, self).__init__()
        
        # Set the parameters specific to Noise
        self.duration = float(duration)
        self.amplitude = float(amplitude)
        try:
            self.channel = int(channel)
        except TypeError:
            self.channel = channel
        
        # Currently only mono or stereo sound is supported
        if self.channel not in [None, 0, 1]:
            raise ValueError(
                "audio channel must be 0, 1, or None, not {}".format(
                self.channel))

        # Initialize the sound itself
        self.init_sound()

    def init_sound(self):
        """Defines `self.table`, the waveform that is played. 
        
        The way this is generated depends on `self.server_type`, because
        parameters like the sampling rate cannot be known otherwise.
        
        The sound is generated and then it is "chunked" (zero-padded and
        divided into chunks). Finally `self.initialized` is set True.
        """
        # Depends on the server_type
        if server_type == 'pyo':
            noiser = pyo.Noise(mul=self.amplitude)
            self.table = self.table_wrap(noiser)
        
        elif server_type == 'jack':
            # This calculates the number of samples, using the specified 
            # duration and the sampling rate from the server, and stores it
            # as `self.nsamples`.
            self.get_nsamples()
            
            # Generate the table by sampling from a uniform distribution
            # The shape of the table depends on `self.channel`
            if self.channel is None:
                # The table will be 1-dimensional for mono sound
                self.table = np.random.uniform(-1, 1, self.nsamples)
            else:
                # The table will be 2-dimensional for stereo sound
                # Each channel is a column
                # Only the specified channel contains data and the other is zero
                data = np.random.uniform(-1, 1, self.nsamples)
                self.table = np.zeros((self.nsamples, 2))
                assert self.channel in [0, 1]
                self.table[:, self.channel] = data
            
            # Scale by the amplitude
            self.table = self.table * self.amplitude
            
            # Convert to float32
            self.table = self.table.astype(np.float32)
            
            # Chunk the sound 
            self.chunk()

        # Flag as initialized
        self.initialized = True

class File(BASE_CLASS):
    """
    A .wav file.

    TODO:
        Generalize this to other audio types if needed.
    """

    PARAMS = ['path', 'amplitude']
    type='File'

    def __init__(self, path, amplitude=0.01, **kwargs):
        """
        Args:
            path (str): Path to a .wav file relative to the `prefs.get('SOUNDDIR')`
            amplitude (float): amplitude of the sound as a proportion of 1.
            **kwargs: extraneous parameters that might come along with instantiating us
        """
        super(File, self).__init__()

        if os.path.exists(path):
            self.path = path
        elif os.path.exists(os.path.join(prefs.get('SOUNDDIR'), path)):
            self.path = os.path.join(prefs.get('SOUNDDIR'), path)
        else:
            Exception('Could not find {} in current directory or sound directory'.format(path))

        self.amplitude = float(amplitude)

        # because files can be v memory intensive, we only load the sound once we're called to buffer them
        # store our initialization status
        self.initialized = False

        #self.init_sound()

    def init_sound(self):
        """
        Load the wavfile with :mod:`scipy.io.wavfile` ,
        converting int to float as needed.

        Create a sound table, resampling sound if needed.
        """

        fs, audio = wavfile.read(self.path)
        if audio.dtype in ['int16', 'int32']:
            audio = int_to_float(audio)

        # load file to sound table
        if self.server_type == 'pyo':
            self.dtable = pyo.DataTable(size=audio.shape[0], chnls=prefs.get('NCHANNELS'), init=audio.tolist())

            # get server to determine sampling rate modification and duration
            server_fs = self.dtable.getServer().getSamplingRate()
            self.duration = float(self.dtable.getSize()) / float(fs)
            self.table = pyo.TableRead(table=self.dtable, freq=float(fs) / server_fs,
                                       loop=False, mul=self.amplitude)

        elif self.server_type == 'jack':
            # attenuate amplitude
            audio = audio*self.amplitude
            self.duration = float(audio.shape[0]) / fs
            # resample to match our audio server's sampling rate
            if fs != self.fs:
                new_samples = self.duration*self.fs
                audio = resample(audio, new_samples)

            self.table = audio

        self.initialized = True

class Gap(BASE_CLASS):
    """
    A silent sound that does not pad its final chunk -- used for creating precise silent
    gaps in a continuous noise.

    """

    type = "Gap"
    PARAMS = ['duration']

    def __init__(self, duration, **kwargs):
        """
        Args:
            duration (float): duration of gap in ms

        Attributes:
            gap_zero (bool): True if duration is zero, effectively do nothing on play.
        """
        super(Gap, self).__init__()

        self.duration = float(duration)
        self.gap_zero = False

        if self.duration == 0:
            self.gap_zero = True
            self.get_nsamples()
            self.chunks = []
            self.table = np.ndarray((0,),dtype=np.float32)
            self.initialized = True
        else:

            self.init_sound()

    def init_sound(self):
        """
        Create and chunk an array of zeros according to :attr:`.Gap.duration`
        """
        if self.server_type == "pyo":
            raise NotImplementedError("This sound has not been implemented for pyo sound server -- pyo is deprecated, and kept as a skeleton in the case interested programmers want to revive its use")

        # get the number of samples for the sound given our self.duration
        self.get_nsamples()
        self.table = np.zeros((self.nsamples,), dtype=np.float32)

        # chunk without padding -- jackclient will pad with ongoing continuous noise (or silence if none)
        self.chunk(pad=False)

        self.initialized = True

    def chunk(self, pad=False):
        """
        If gap is not duration == 0, call parent ``chunk``.
        Args:
            pad (bool): unused, passed to parent ``chunk``
        """
        if not self.gap_zero:
            super(Gap, self).chunk(pad)
        else:
            self.padded=False


    def buffer(self):
        if not self.gap_zero:
            super(Gap, self).buffer()
        else:
            self.buffered = True

    def play(self):
        if not self.gap_zero:
            super(Gap, self).play()
        else:
            if callable(self.trigger):
                threading.Thread(target=self.wait_trigger).start()


class Gammatone(Noise):
    """
    Gammatone filtered noise, using :class:`.timeseries.Gammatone` --
    see that class for the filter documentation.
    """

    type = "Gammatone"

    PARAMS = Noise.PARAMS.copy()
    PARAMS.insert(0, 'frequency')

    def __init__(self,
                 frequency:float, duration:float, amplitude:float=0.01,
                 channel:typing.Optional[int]=None,
                 **kwargs):
        """
        Args:
            frequency (float): Center frequency of filter, in Hz
            duration (float): Duration of sound, in ms
            amplitude (float): Amplitude scaling of sound (absolute value 0-1, default is .01)
            **kwargs: passed on to :class:`.timeseries.Gammatone`
        """

        super(Gammatone, self).__init__(duration, amplitude, channel, **kwargs)

        self.frequency = float(frequency)
        self.kwargs = kwargs

        self.filter = autopilot.get('transform', 'Gammatone')(
            self.frequency, self.fs, axis=0, **self.kwargs
        )

        # superclass init calls its init sound, so we just call the gammatone filter part
        self._init_sound()

    def init_sound(self):
        # superclass generates the noise table and chunks it, so
        # we need to apply the filter to the table and rechunk
        super(Gammatone, self).init_sound()
        self._init_sound()

    def _init_sound(self):
        # just the gammatone specific parts so they can be called separately on init
        self.table = self.filter.process(self.table)
        self.chunk()





    # These parameters are strings not numbers... jonny should do this better
STRING_PARAMS = ['path', 'type']
"""
These parameters should be given string columns rather than float columns.

Bother Jonny to do this better bc it's really bad.
"""


## Helper function
def int_to_float(audio):
    """
    Convert 16 or 32 bit integer audio to 32 bit float.

    Args:
        audio (:class:`numpy.ndarray`): a numpy array of audio

    Returns:
        :class:`numpy.ndarray`: Audio that has been rescaled and converted to a 32 bit float.
    """
    if audio.dtype == 'int16':
        audio = audio.astype(np.float32)
        audio = audio / (float(2 ** 16) / 2)
    elif audio.dtype == 'int32':
        audio = audio.astype(np.float32)
        audio = audio / (float(2 ** 32) / 2)

    return audio



import queue
import threading
import time
import typing
from copy import copy
from itertools import count
from typing import Union, Optional
from collections import deque
import socket

import zmq
from tornado.ioloop import IOLoop
from zmq.eventloop.zmqstream import ZMQStream

from autopilot import prefs
from autopilot.core.loggers import init_logger
from autopilot.networking.message import Message


class Net_Node(object):
    """
    Drop in networking object to be given to any sub-object
    behind some external-facing :class:`.Station` object.

    To minimize the complexity of the network topology, the typical way to use
     ``Net_Node``s is through a :class:`.Station` ROUTER, rather than
    addressing each other directly. Practically, this means that
    all messages are sent first to the parent :class:`.networking.Station` object,
    which then handles them, forwards them, etc.
    This proved to be horribly misguided and
    will be changed in v0.5.0 to support simplified
    messaging to a ``agent_id.netnode_id`` address. Until then the networking modules
    will be in a bit of flux.

    To receive messages directly at this Net_Node, pass the ``router_port``
    which will bind a ``zmq.ROUTER`` socket, and messages will be handled as regular 'listens'
    Note that Net_Nodes assume that they are the final recipients of messages,
    and so don't handle forwarding messages (unless a ``listen`` method explicitly
    does so), and will automatically deserialize them on receipt.

    .. note::

        Listen methods currently receive only the ``value`` of a message, this will change in v0.5.0,
        where they will receive the full message like :class:`.networking.Station` objects.

    Args:
        id (str): What are we known as? What do we set our :attr:`~zmq.Socket.identity` as?
        upstream (str): The identity of the ROUTER socket used by our upstream :class:`.Station` object.
        port (int): The port that our upstream ROUTER socket is bound to
        listens (dict): Dictionary of functions to call for different types of messages.
            keys match the :attr:`.Message.key`.
        instance (bool): Should the node try and use the existing zmq context and tornado loop?
        upstream_ip (str): If this Net_Node is being used on its own (ie. not behind a :class:`.Station`), it can directly connect to another node at this IP. Otherwise use 'localhost' to connect to a station.
        router_port (int): Typically, Net_Nodes only have a single Dealer socket and receive messages from their encapsulating :class:`.Station`, but
            if you want to take this node offroad and use it independently, an int here binds a Router to the port.
        daemon (bool): Run the IOLoop thread as a ``daemon`` (default: ``True``)

    Attributes:
        context (:class:`zmq.Context`):  zeromq context
        loop (:class:`tornado.ioloop.IOLoop`): a tornado ioloop
        sock (:class:`zmq.Socket`): Our DEALER socket.
        id (str): What are we known as? What do we set our :attr:`~zmq.Socket.identity` as?
        upstream (str): The identity of the ROUTER socket used by our upstream :class:`.Station` object.
        port (int): The port that our upstream ROUTER socket is bound to
        listens (dict): Dictionary of functions to call for different types of messages. keys match the :attr:`.Message.key`.
        outbox (dict): Messages that have been sent but have not been confirmed
        timers (dict): dict of :class:`threading.Timer` s that will check in on outbox messages
        logger (:class:`logging.Logger`): Used to log messages and network events.
        msg_counter (:class:`itertools.count`): counter to index our sent messages
        loop_thread (:class:`threading.Thread`): Thread that holds our loop. initialized with `daemon=True`
    """
    repeat_interval = 5 # how many seconds to wait before trying to repeat a message

    def __init__(self, id: str, upstream: str, port: int,
                 listens: typing.Dict[str, typing.Callable],
                 instance:bool=True, upstream_ip:str='localhost',
                 router_port:Optional[int] = None,
                 daemon:bool=True, expand_on_receive:bool=True):

        if instance:
            self.context = zmq.Context.instance() # type: zmq.Context
            self.loop    = IOLoop.current() # type: IOLoop
        else:
            self.context = zmq.Context()  # type: zmq.Context
            self.loop    = IOLoop() # type: IOLoop

        self.closing = threading.Event() # type: threading.Event
        self.closing.clear()

        # we have a few builtin listens
        self.listens = {
            'CONFIRM': self.l_confirm,
            #'STREAM' : self.l_stream
        } # type: typing.Dict[str, typing.Callable]
        # then add the rest
        self.listens.update(listens)

        self.id = id # type: str
        self.upstream = upstream  # type: str
        self.port = int(port)  # type: int
        self.router_port = router_port
        self.router = None # type: Optional[zmq.Socket]
        self.loop_thread = None  # type: Optional[threading.Thread]
        self.senders = {} # type: typing.Dict[bytes, str]
        self._ip = None

        # self.connected = False
        self.msg_counter = count()
        self.msgs_received = 0
        self.logger = init_logger(self)

        # If we were given an explicit IP to connect to, stash it
        self.upstream_ip = upstream_ip

        self.daemon = daemon
        self.streams = {}
        self.outbox = {}
        self.timers = {}
        self.expand = expand_on_receive

        if prefs.get( 'SUBJECT'):
            self.subject = prefs.get('SUBJECT').encode('utf-8')
        else:
            self.subject = None

        self.init_networking()

    def __del__(self):
        self.release()

    def init_networking(self):
        """
        Creates socket, connects to specified port on localhost,
        and starts the :meth:`~Net_Node.threaded_loop` as a daemon thread.
        """
        self.sock = self.context.socket(zmq.DEALER)
        self.sock.setsockopt_string(zmq.IDENTITY, self.id)
        #self.sock.probe_router = 1

        # connect our dealer socket to "push" messages upstream
        self.sock.connect('tcp://{}:{}'.format(self.upstream_ip, self.port))
        self.sock = ZMQStream(self.sock, self.loop)
        self.sock.on_recv(self.handle_listen)

        # if want to directly receive messages, bind a router port
        if self.router_port is not None:
            self.router = self.context.socket(zmq.ROUTER)
            self.router.setsockopt_string(zmq.IDENTITY, self.id)
            self.router.bind('tcp://*:{}'.format(self.router_port))
            self.router = ZMQStream(self.router, self.loop)
            self.router.on_recv(self.handle_listen)

        # the loop thread keeps the ioloop alive until the program exits
        self.loop_thread = threading.Thread(target=self.threaded_loop)
        if self.daemon:
            self.loop_thread.daemon = True
        self.loop_thread.start()

    def threaded_loop(self):
        """
        Run in a thread, either starts the IOLoop, or if it
        is already started (ie. running in another thread),
        breaks.
        """

        while not self.closing.is_set():
            try:
                self.loop.start()
            except RuntimeError:
                # loop already started
                break

    def handle_listen(self, msg: typing.List[bytes]):
        """
        Upon receiving a message, call the appropriate listen method
        in a new thread and send confirmation it was received.

        Note:
            Unlike :meth:`.Station.handle_listen` , only the :attr:`.Message.value`
            is given to listen methods. This was initially intended to simplify these
            methods, but this might change in the future to unify the messaging system.

        Args:
            msg (list): JSON :meth:`.Message.serialize` d message.
        """
        self.msgs_received += 1

        # if we have a router, check if this is a router msg and store
        # the sender if so
        if self.router is not None and len(msg)>=2:
            if msg[0] not in self.senders.keys():
                self.senders[msg[0]] = ''

        # Nodes expand arrays by default as they're expected to
        msg = Message(msg[-1], expand_arrays=self.expand)

        # Check if our listen was sent properly
        if not msg.validate():
            if self.logger:
                self.logger.error('Message failed to validate:\n{}'.format(str(msg)))
            return

        # unnest any list if it was a multihop message
        if isinstance(msg.to, list) and len(msg.to) == 1:
            msg.to = msg.to[0]

        try:
            listen_funk = self.listens[msg.key]
            listen_thread = threading.Thread(target=listen_funk, args=(msg.value,))
            listen_thread.start()
        except KeyError:
            if msg.key=="STREAM":
                try:
                    listen_thread = threading.Thread(target=self.l_stream, args=(msg,))
                    listen_thread.start()
                except Exception as e:
                    self.logger.exception(e)

            self.logger.exception('MSG ID {} - No listen function found for key: {}'.format(msg.id, msg.key))

        if (msg.key != "CONFIRM") and ('NOREPEAT' not in msg.flags.keys()) :
            # send confirmation
            self.send(msg.sender, 'CONFIRM', msg.id)

        log_this = True
        if 'NOLOG' in msg.flags.keys():
            log_this = False

        if self.logger and log_this:
            self.logger.debug('RECEIVED: {}'.format(str(msg)))


    def send(self, to: Optional[Union[str, list]] = None,
             key:str=None,
             value:typing.Any=None,
             msg:Optional['Message']=None,
             repeat:bool=True, flags = None, force_to:bool = False):
        """
        Send a message via our :attr:`~.Net_Node.sock` , DEALER socket.

        `to` is not required.

        * If the node doesn't have a router, (or the recipient is not
          in the :attr:`Net_Node.senders` dict ) every message
          is always sent to :attr:`~.Net_Node.upstream` . `to` can be included
          to send a message further up the network tree to a networking object
          we're not directly connected to.
        * If the node has a router, since messages can only be sent on router
          sockets after the recipient has first sent us a message, if the
          ``to`` is in the :attr:`~.Net_Node.senders` dict, it will be
          directly sent via :attr:`.Net_Node.router`
        * If the ``force_to`` arg is ``True``, send to the ``to`` recipient directly
          via the dealer :attr:`.Net_Node.sock`
        * If ``to`` is a list, or is intended to be sent as a multihop message with
          an explicit path, then networking objects will attempt to forward it
          along that path (disregarding implicit topology).

        Either an already created :class:`.Message` should be passed as `msg`,
        or at least `key` must be provided for a new message created
        by :meth:`~.Net_Node.prepare_message` .

        A :class:`threading.Timer` is created to resend the message using
        :meth:`~.Net_Node.repeat` unless `repeat` is False.

        Args:
            to (str, list): The identity of the socket this message is to. If not included,
                sent to :meth:`~.Net_Node.upstream` .
            key (str): The type of message - used to select which method the receiver
                uses to process this message.
            value: Any information this message should contain. Can be any type, but
                must be JSON serializable.
            msg (`.Message`): An already created message.
            repeat (bool): Should this message be resent if confirmation is not received?
            flags (dict):
            force_to (bool): If we really really want to use the 'to' field to address messages
                (eg. node being used for direct communication), overrides default behavior of sending to upstream.
        """
        if (key is None) and (msg is None):
            if self.logger:
                self.logger.error('Push sent without Key')
            return

        # send message via the dealer
        # even though we only have one connection over our dealer,
        # we still include 'to' in case we are sending further upstream
        # but can push without 'to', just fill in with upstream id
        if to is None:
            to = self.upstream

        # differentiate between a single 'to' and a list (ie. a multihop message)
        # in this case, 'recipient' is encoded in the message as the final node to
        # send to, and the rest of 'to' is encoded as parts in a multipart message.
        if isinstance(to, list):
            recipient = to[-1]
        else:
            recipient = to

        if not msg:
            msg = self.prepare_message(recipient, key, value, repeat, flags)

        log_this = True
        if 'NOLOG' in msg.flags.keys():
            log_this = False

        # encode message
        msg_enc = msg.serialize()
        if not msg_enc:
            self.logger.error('Message could not be encoded:\n{}'.format(str(msg)))
            return

        if isinstance(to, list):
            multipart = [bytes(hop, encoding='utf-8') for hop in to]
            multipart.append(recipient.encode('utf-8'))
            multipart.append(msg_enc)

        else:
            # the first frame will be added below if needed...
            multipart = [recipient.encode('utf-8'), msg_enc]
            if force_to or to.encode('utf-8') in self.senders.keys():
                multipart.insert(0, to.encode('utf-8'))
            else:
                multipart.insert(0, self.upstream.encode('utf-8'))

        if self.router is not None and multipart[0] in self.senders.keys():
            self.router.send_multipart(multipart)
        else:
            self.sock.send_multipart(multipart)

        if self.logger and log_this:
            self.logger.debug("MESSAGE SENT - {}".format(str(msg)))

        if repeat and not msg.key == "CONFIRM":
            # add to outbox and spawn timer to resend
            self.outbox[msg.id] = (time.time(), msg)

    def repeat(self):
        """
        Periodically (according to :attr:`~.repeat_interval`) resend messages that haven't been confirmed

        TTL is decremented, and messages are resent until their TTL is 0.

        """
        while not self.closing.is_set():
            # try to send any outstanding messages and delete if too old
            # make a local copy of dict
            outbox = copy(self.outbox)

            if len(outbox) > 0:
                for id in outbox.keys():
                    if outbox[id][1].ttl <= 0:
                        self.logger.warning('PUBLISH FAILED {} - {}'.format(id, str(outbox[id][1])))
                        try:
                            del self.outbox[id]
                        except KeyError:
                            # fine, already deleted
                            pass
                    else:
                        # if we didn't just put this message in the outbox...
                        if (time.time() - outbox[id][0]) > (self.repeat_interval*2):
                            self.logger.debug('REPUBLISH {} - {}'.format(id, str(outbox[id][1])))
                            self.sock.send_multipart([self.upstream.encode('utf-8'), outbox[id][1].serialize()])
                            self.outbox[id][1].ttl -= 1


            # wait to do it again
            time.sleep(self.repeat_interval)

    def l_confirm(self, value):
        """
        Confirm that a message was received.

        Args:
            value (str): The ID of the message we are confirming.
        """
        # delete message from outbox if we still have it
        # msg.value should contain the if of the message that was confirmed
        try:
            if value in self.outbox.keys():
                del self.outbox[value]
        except KeyError:
            # already deleted
            pass

        # # stop a timer thread if we have it
        # if value in self.timers.keys():
        #     self.timers[value].cancel()
        #     del self.timers[value]


        self.logger.debug('CONFIRMED MESSAGE {}'.format(value))

    def l_stream(self, msg):
        """
        Reconstitute the original stream of messages and call their handling methods

        The ``msg`` should contain an ``inner_key`` that indicates the key, and thus the
        handling method.

        Args:
            msg (dict): Compressed stream sent by :meth:`Net_Node._stream`
        """
        listen_fn = self.listens[msg.value['inner_key']]
        old_value = copy(msg.value)
        delattr(msg, 'value')
        for v in old_value['payload']:
            # if isinstance(v, dict) and ('headers' in old_value.keys()):
            #     v.update(old_value['headers'])
            #msg.value = v
            listen_fn(v)
    #
    # def l_stream(self, value):
    #     listen_fn = self.listens[value['inner_key']]
    #     for v in value['payload']:
    #         listen_fn(v)
    #
    #


    def prepare_message(self, to, key, value, repeat, flags=None):
        """
        Instantiate a :class:`.Message` class, give it an ID and
        the rest of its attributes.

        Args:
            flags:
            repeat:
            to (str): The identity of the socket this message is to
            key (str): The type of message - used to select which method the receiver
                uses to process this message.
            value: Any information this message should contain. Can be any type, but
                must be JSON serializable.
        """
        msg = Message()

        # if our name is _{something} and our upstream is {something}, replace sender with our upstream node
        # upstream node should handle all incoming information to those types of nodes
        #if self.id == "_{}".format(self.upstream):
        #    msg.sender = self.upstream
        #else:
        msg.sender = self.id

        try:
            msg.to = to.decode('utf-8')
        except AttributeError:
            msg.to = to

        try:
            msg.key = key.decode('utf-8')
        except AttributeError:
            msg.key = key

        msg.value = value

        msg_num = next(self.msg_counter)
        msg.id = "{}_{}".format(self.id, msg_num)

        if not repeat:
            msg.flags['NOREPEAT'] = True


        if flags:
            for k, v in flags.items():
                msg.flags[k] = v


        return msg

    def get_stream(self, id, key, min_size=5, upstream=None, port = None, ip=None, subject=None, q_size:Optional[int]=None):
        """

        Make a queue that another object can dump data into that sends on its own socket.
        Smarter handling of continuous data than just hitting 'send' a shitload of times.
        Returns:
            Queue: Place to dump ur data

        """
        if upstream is None:
            upstream = self.upstream

        if port is None:
            port = self.port

        if ip is None:
            ip = self.upstream_ip

        if subject is None:
            if self.subject:
                subject = self.subject
            elif prefs.get( 'SUBJECT'):
                subject = prefs.get('SUBJECT')

        # make a queue
        q = deque(maxlen=q_size)

        stream_thread = threading.Thread(target=self._stream,
                                         args=(id, key, min_size, upstream, port, ip, subject, q))
        stream_thread.setDaemon(True)
        stream_thread.start()
        self.streams[id] = stream_thread

        self.streams[id] = stream_thread

        self.logger.info(("Stream started with configuration:\n"+
                          "ID: {}\n".format(self.id+"_"+id)+
                          "Key: {}\n".format(key)+
                          "Min Chunk Size: {}\n".format(min_size)+
                          "Upstream ID: {}\n".format(upstream) +
                          "Port: {}\n".format(port) +
                          "IP: {}\n".format(ip) +
                          "Subject: {}\n".format(subject)))



        return q


    def _stream(self, id, msg_key, min_size, upstream, port, ip, subject, q):



        # create a new context and socket
        #context = zmq.Context()
        #loop = IOLoop()
        socket = self.context.socket(zmq.DEALER)
        socket_id = "{}_{}".format(self.id, id)
        #socket.identity = socket_id
        socket.setsockopt_string(zmq.IDENTITY, socket_id)
        socket.connect('tcp://{}:{}'.format(ip, port))

        socket = ZMQStream(socket, self.loop)

        upstream = upstream.encode('utf-8')

        if subject is None:
            if prefs.get( 'SUBJECT'):
                subject = prefs.get('SUBJECT')
            else:
                subject = ""
        if isinstance(subject, bytes):
            subject = subject.decode('utf-8')

        if prefs.get('LINEAGE') == "CHILD":
            # pilot = bytes(prefs.get('PARENTID'), encoding="utf-8")
            pilot = prefs.get('PARENTID')
        else:
            # pilot = bytes(prefs.get('NAME'), encoding="utf-8")
            pilot = prefs.get('NAME')

        msg_counter = count()

        pending_data = []

        if min_size > 1:

            while True:
                try:
                    data = q.popleft()
                except IndexError:
                    # normal, we might iterate faster than the source
                    continue
                if isinstance(data, str) and data == 'END':
                    break
                if isinstance(data, tuple):
                    # tuples are immutable, so can't serialize numpy arrays they contain
                    data = list(data)

                pending_data.append(data)

                if not socket.sending() and len(pending_data)>=min_size:
                    msg = Message(to=upstream.decode('utf-8'), key="STREAM",
                                  value={'inner_key' : msg_key,
                                         'headers'   : {'subject': subject,
                                                        'pilot'  : pilot,
                                                        'continuous': True},
                                         'payload'   : pending_data},
                                  id="{}_{}".format(id, next(msg_counter)),
                                  flags={'NOREPEAT':True, 'MINPRINT':True},
                                  sender=socket_id).serialize()
                    last_msg = socket.send_multipart((upstream, upstream, msg),
                                                     track=True, copy=True)

                    self.logger.debug("STREAM {}: Sent {} items".format(self.id+'_'+id, len(pending_data)))
                    pending_data = []
        else:
            # just send like normal messags
            # just send like normal messags
            while True:
                try:
                    data = q.popleft()
                except IndexError:
                    continue

                if isinstance(data, str) and data == "END":
                    break

                if isinstance(data, tuple):
                    # tuples are immutable, so can't serialize numpy arrays they contain
                    data = list(data)

                msg = Message(to=upstream.decode('utf-8'), key=msg_key,
                              subject=subject,
                              pilot=pilot,
                              continuous=True,
                              value=data,
                              flags={'NOREPEAT': True, 'MINPRINT': True},
                              id="{}_{}".format(id, next(msg_counter)),
                              sender=socket_id).serialize()
                socket.send_multipart((upstream, upstream, msg),
                                       track=False, copy=False)

                self.logger.debug("STREAM {}: Sent 1 item".format(self.id + '_' + id))


    @property
    def ip(self) -> str:
        """
        Find our IP address

        .. todo::

            this is a copy of the :meth:`.Station.get_ip` method -- unify this in v0.5.0

        returns (str): our IPv4 address.
        """

        # shamelessly stolen from https://www.w3resource.com/python-exercises/python-basic-exercise-55.php
        # variables are badly named because this is just a rough unwrapping of what was a monstrous one-liner
        # (and i don't really understand how it works)

        if self._ip is None:

            # get ips that aren't the loopback
            unwrap00 = [ip for ip in socket.gethostbyname_ex(socket.gethostname())[2] if not ip.startswith("127.")][:1]
            # ??? truly dk
            unwrap01 = [[(s.connect(('8.8.8.8', 53)), s.getsockname()[0], s.close()) for s in
                         [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]]

            self._ip = [l for l in (unwrap00, unwrap01) if l][0][0]

        return self._ip

    def release(self):
        self.closing.set()
        self.loop.stop()